# Copyright 2006-2009 Divmod, Inc.  See LICENSE file for details

"""
Tests for L{axiom.scripts.axiomatic}.
"""

import sys, os, signal, StringIO

from zope.interface import implements

from twisted.python.log import msg
from twisted.python.filepath import FilePath
from twisted.python.procutils import which
from twisted.python.runtime import platform
from twisted.trial.unittest import SkipTest, TestCase
from twisted.plugin import IPlugin
from twisted.internet import reactor
from twisted.internet.task import deferLater
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.defer import Deferred
from twisted.internet.error import ProcessTerminated
from twisted.application.service import IService, IServiceCollection

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import boolean
from axiom.scripts import axiomatic
from axiom.listversions import SystemVersion
from axiom.iaxiom import IAxiomaticCommand
from twisted.plugins.axiom_plugins import AxiomaticStart

from axiom.test.reactorimporthelper import SomeItem


class RecorderService(Item):
    """
    Minimal L{IService} implementation which remembers if it was ever started.
    This is used by tests to make sure services get started when they should
    be.
    """
    implements(IService)

    started = boolean(
        doc="""
        A flag which is initially false and set to true once C{startService} is
        called.
        """, default=False)

    name = "recorder"

    def setServiceParent(self, parent):
        """
        Do the standard Axiom thing to make sure this service becomes a child
        of the top-level store service.
        """
        IServiceCollection(parent).addService(self)


    def startService(self):
        """
        Remember that this method was called.
        """
        self.started = True


    def stopService(self):
        """
        Ignore this event.
        """



class StartTests(TestCase):
    """
    Test the axiomatic start sub-command.
    """
    maxDiff = None
    def setUp(self):
        """
        Work around Twisted #3178 by tricking trial into thinking something
        asynchronous is happening.
        """
        return deferLater(reactor, 0, lambda: None)


    def _getRunDir(self, dbdir):
        return dbdir.child("run")


    def _getLogDir(self, dbdir):
        return self._getRunDir(dbdir).child("logs")


    def test_getArguments(self):
        """
        L{Start.getArguments} adds a I{--pidfile} argument if one is not
        present and a I{--logfile} argument if one is not present and
        daemonization is enabled and adds a I{--dbdir} argument pointing at the
        store it is passed. It also adds I{--journal-mode} if this is passed.
        """
        dbdir = FilePath(self.mktemp())
        store = Store(dbdir, journalMode=u'WAL')
        run = self._getRunDir(dbdir)
        logs = self._getLogDir(dbdir)
        start = axiomatic.Start()

        logfileArg = ["--logfile", logs.child("axiomatic.log").path]

        # twistd on Windows doesn't support PID files, so on Windows,
        # getArguments should *not* add --pidfile.
        if platform.isWindows():
            pidfileArg = []
        else:
            pidfileArg = ["--pidfile", run.child("axiomatic.pid").path]
        restArg = [
            "axiomatic-start", "--dbdir", dbdir.path, "--journal-mode", "WAL"]

        self.assertEqual(
            start.getArguments(store, []),
            logfileArg + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["--logfile=foo"]),
            ["--logfile=foo"] + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["--logfile", "foo"]),
            ["--logfile", "foo"] + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["-l", "foo"]),
            ["-l", "foo"] + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["--nodaemon"]),
            ["--nodaemon"] + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["-n"]),
            ["-n"] + pidfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["--pidfile=foo"]),
            ["--pidfile=foo"] + logfileArg + restArg)
        self.assertEqual(
            start.getArguments(store, ["--pidfile", "foo"]),
            ["--pidfile", "foo"] + logfileArg + restArg)


    def test_logDirectoryCreated(self):
        """
        If L{Start.getArguments} adds a I{--logfile} argument, it creates the
        necessary directory.
        """
        dbdir = FilePath(self.mktemp())
        store = Store(dbdir)
        start = axiomatic.Start()
        start.getArguments(store, ["-l", "foo"])
        self.assertFalse(self._getLogDir(dbdir).exists())
        start.getArguments(store, [])
        self.assertTrue(self._getLogDir(dbdir).exists())


    def test_parseOptions(self):
        """
        L{Start.parseOptions} adds axiomatic-suitable defaults for any
        unspecified parameters and then calls L{twistd.run} with the modified
        argument list.
        """
        argv = []
        def fakeRun():
            argv.extend(sys.argv)
        options = axiomatic.Options()
        options['dbdir'] = dbdir = self.mktemp()
        start = axiomatic.Start()
        start.parent = options
        start.run = fakeRun
        original = sys.argv[:]
        try:
            start.parseOptions(["-l", "foo", "--pidfile", "bar"])
        finally:
            sys.argv[:] = original
        self.assertEqual(
            argv,
            [sys.argv[0],
             "-l", "foo", "--pidfile", "bar",
             "axiomatic-start", "--dbdir", os.path.abspath(dbdir)])


    def test_parseOptionsHelp(self):
        """
        L{Start.parseOptions} writes usage information to stdout if C{"--help"}
        is in the argument list it is passed and L{twistd.run} is not called.
        """
        start = axiomatic.Start()
        start.run = None
        original = sys.stdout
        sys.stdout = stdout = StringIO.StringIO()
        try:
            self.assertRaises(SystemExit, start.parseOptions, ["--help"])
        finally:
            sys.stdout = original

        # Some random options that should be present.  This is a bad test
        # because we don't control what C{opt_help} actually does and we don't
        # even really care as long as it's the same as what I{twistd --help}
        # does.  We could try running them both and comparing, but then we'd
        # still want to do some sanity check against one of them in case we end
        # up getting the twistd version incorrectly somehow... -exarkun
        self.assertIn("--reactor", stdout.getvalue())
        if not platform.isWindows():
            # This isn't an option on Windows, so it shouldn't be there.
            self.assertIn("--uid", stdout.getvalue())

        # Also, we don't want to see twistd plugins here.
        self.assertNotIn("axiomatic-start", stdout.getvalue())



    def test_checkSystemVersion(self):
        """
        The L{IService} returned by L{AxiomaticStart.makeService} calls
        L{checkSystemVersion} with its store when it is started.

        This is done for I{axiomatic start} rather than somewhere in the
        implementation of L{Store} so that it happens only once per server
        startup.  The overhead of doing it whenever a store is opened is
        non-trivial.
        """
        dbdir = self.mktemp()
        store = Store(dbdir)
        service = AxiomaticStart.makeService(
            {'dbdir': dbdir, 'debug': False, 'journal-mode': None})
        self.assertEqual(store.query(SystemVersion).count(), 0)
        service.startService()
        self.assertEqual(store.query(SystemVersion).count(), 1)
        return service.stopService()


    def test_axiomOptions(self):
        """
        L{AxiomaticStart.options} takes database location, debug, and
        journal mode setting parameters.
        """
        options = AxiomaticStart.options()
        options.parseOptions([])
        self.assertEqual(options['dbdir'], None)
        self.assertFalse(options['debug'])
        self.assertEqual(options['journal-mode'], None)
        options.parseOptions(
            ["--dbdir", "foo", "--debug", "--journal-mode", "WAL"])
        self.assertEqual(options['dbdir'], 'foo')
        self.assertTrue(options['debug'])
        self.assertEqual(options['journal-mode'], 'WAL')


    def test_makeService(self):
        """
        L{AxiomaticStart.makeService} returns the L{IService} powerup of the
        L{Store} at the directory in the options object it is passed.
        """
        dbdir = FilePath(self.mktemp())
        store = Store(dbdir)
        recorder = RecorderService(store=store)
        self.assertFalse(recorder.started)
        store.powerUp(recorder, IService)
        store.close()

        service = AxiomaticStart.makeService(
            {'dbdir': dbdir, 'debug': False, 'journal-mode': None})
        service.startService()
        service.stopService()

        store = Store(dbdir)
        self.assertTrue(store.getItemByID(recorder.storeID).started)


    def _getAxiomaticScript(self):
        here = FilePath(__file__)
        # Try to find it relative to the source of this test.
        bin = here.parent().parent().parent().child("bin")
        axiomatic = bin.child("axiomatic")
        if axiomatic.exists():
            # Great, use that one.
            axiomatic = axiomatic.path
        else:
            # Try to find it on the path, instead.
            axiomatics = which("axiomatic")
            if axiomatics:
                # Great, it was on the path.
                axiomatic = axiomatics[0]
            else:
                # Nope, not there, give up.
                raise SkipTest(
                    "Could not find axiomatic script on path or at %s" % (
                        axiomatic.path,))
        return axiomatic


    def test_reactorSelection(self):
        """
        L{AxiomaticStart} optionally takes the name of a reactor in the form
        --reactor [shortName] and installs it instead of the default reactor.
        """
        # Since this process is already hopelessly distant from the state in
        # which I{axiomatic start} operates, it would make no sense to try a
        # functional test of this behavior in this process.  Since the
        # behavior being tested involves lots of subtle interactions between
        # lots of different pieces of code (the reactor might get installed
        # at the end of a ten-deep chain of imports going through as many
        # different projects), it also makes no sense to try to make this a
        # unit test.  So, start a child process and try to use the alternate
        # reactor functionality there.

        axiomatic = self._getAxiomaticScript()

        # Create a store for the child process to use and put an item in it.
        # This will force an import of the module that defines that item's
        # class when the child process starts.  The module imports the default
        # reactor at the top-level, making this the worst-case for the reactor
        # selection code.
        storePath = self.mktemp()
        store = Store(storePath)
        SomeItem(store=store)
        store.close()

        # Install select reactor because it is available on all platforms, and
        # it is still an error to try to install the select reactor even if
        # the already installed reactor was the select reactor.
        argv = [
            sys.executable,
            axiomatic, "-d", storePath,
            "start", "--reactor", "select", "-n"]
        expected = [
            "reactor class: twisted.internet.selectreactor.SelectReactor.",
            "reactor class: <class 'twisted.internet.selectreactor.SelectReactor'>"]
        proto, complete = AxiomaticStartProcessProtocol.protocolAndDeferred(expected)

        environ = os.environ.copy()
        reactor.spawnProcess(proto, sys.executable, argv, env=environ)
        return complete


    def test_reactorSelectionLongOptionStyle(self):
        """
        L{AxiomaticStart} optionally takes the name of a reactor in the form
        --reactor=[shortName] and installs it instead of the default reactor.
        """

        axiomatic = self._getAxiomaticScript()

        # Create a store for the child process to use and put an item in it.
        # This will force an import of the module that defines that item's
        # class when the child process starts.  The module imports the default
        # reactor at the top-level, making this the worst-case for the reactor
        # selection code.
        storePath = self.mktemp()
        store = Store(storePath)
        SomeItem(store=store)
        store.close()

        # Install select reactor because it is available on all platforms, and
        # it is still an error to try to install the select reactor even if
        # the already installed reactor was the select reactor.
        argv = [
            sys.executable,
            axiomatic, "-d", storePath,
            "start", "--reactor=select", "-n"]
        expected = [
            "reactor class: twisted.internet.selectreactor.SelectReactor.",
            "reactor class: <class 'twisted.internet.selectreactor.SelectReactor'>"]
        proto, complete = AxiomaticStartProcessProtocol.protocolAndDeferred(expected)

        environ = os.environ.copy()
        reactor.spawnProcess(proto, sys.executable, argv, env=environ)
        return complete


    def test_reactorSelectionShortOptionStyle(self):
        """
        L{AxiomaticStart} optionally takes the name of a reactor in the form
        -r [shortName] and installs it instead of the default reactor.
        """

        axiomatic = self._getAxiomaticScript()

        # Create a store for the child process to use and put an item in it.
        # This will force an import of the module that defines that item's
        # class when the child process starts.  The module imports the default
        # reactor at the top-level, making this the worst-case for the reactor
        # selection code.
        storePath = self.mktemp()
        store = Store(storePath)
        SomeItem(store=store)
        store.close()

        # Install select reactor because it is available on all platforms, and
        # it is still an error to try to install the select reactor even if
        # the already installed reactor was the select reactor.
        argv = [
            sys.executable,
            axiomatic, "-d", storePath,
            "start", "-r", "select", "-n"]
        expected = [
            "reactor class: twisted.internet.selectreactor.SelectReactor.",
            "reactor class: <class 'twisted.internet.selectreactor.SelectReactor'>"]
        proto, complete = AxiomaticStartProcessProtocol.protocolAndDeferred(expected)

        environ = os.environ.copy()
        reactor.spawnProcess(proto, sys.executable, argv, env=environ)
        return complete



class AxiomaticStartProcessProtocol(ProcessProtocol):
    """
    L{AxiomaticStartProcessProtocol} watches an I{axiomatic start} process
    and fires a L{Deferred} when it sees either successful reactor
    installation or process termination.

    @ivar _success: A flag which is C{False} until the expected text is found
        in the child's stdout and C{True} thereafter.

    @ivar _output: A C{str} giving all of the stdout from the child received
        thus far.
    """
    _success = False
    _output = ""


    def protocolAndDeferred(cls, expected):
        """
        Create and return an L{AxiomaticStartProcessProtocol} and a
        L{Deferred}.  The L{Deferred} will fire when the protocol receives
        the given string on standard out or when the process ends, whichever
        comes first.
        """
        proto = cls()
        proto._complete = Deferred()
        proto._expected = expected
        return proto, proto._complete
    protocolAndDeferred = classmethod(protocolAndDeferred)


    def errReceived(self, bytes):
        """
        Report the given unexpected stderr data.
        """
        msg("Received stderr from axiomatic: %r" % (bytes,))


    def outReceived(self, bytes):
        """
        Add the given bytes to the output buffer and check to see if the
        reactor has been installed successfully, firing the completion
        L{Deferred} if so.
        """
        msg("Received stdout from axiomatic: %r" % (bytes,))
        self._output += bytes
        if not self._success:
            for line in self._output.splitlines():
                for expectedLine in self._expected:
                    if expectedLine in line:
                        msg("Received expected output")
                        self._success = True
                        self.transport.signalProcess("TERM")


    def processEnded(self, reason):
        """
        Check that the process exited in the way expected and that the required
        text has been found in its output and fire the result L{Deferred} with
        either a value or a failure.
        """
        self._complete, result = None, self._complete
        if self._success:
            if platform.isWindows() or (
                # Windows can't tell that we SIGTERM'd it, so sorry.
                reason.check(ProcessTerminated) and
                reason.value.signal == signal.SIGTERM):
                result.callback(None)
                return
        # Something went wrong.
        result.errback(reason)



class TestMisc(TestCase):
    """
    Test things not directly involving running axiomatic commands.
    """
    def test_axiomaticCommandProvides(self):
        """
        Test that AxiomaticCommand itself does not provide IAxiomaticCommand or
        IPlugin, but subclasses do.
        """
        self.failIf(IAxiomaticCommand.providedBy(axiomatic.AxiomaticCommand), 'IAxiomaticCommand provided')
        self.failIf(IPlugin.providedBy(axiomatic.AxiomaticCommand), 'IPlugin provided')

        class _TestSubClass(axiomatic.AxiomaticCommand):
            pass

        self.failUnless(IAxiomaticCommand.providedBy(_TestSubClass), 'IAxiomaticCommand not provided')
        self.failUnless(IPlugin.providedBy(_TestSubClass), 'IPlugin not provided')



class AxiomaticTests(TestCase):
    """
    Test things relating to the I{axiomatic} command itself.
    """
    def test_journalMode(self):
        """
        I{axiomatic} sets the journal mode of the store according to
        I{--journal-mode}.
        """
        options = axiomatic.Options()
        options['dbdir'] = self.mktemp()
        options['journal-mode'] = 'WAL'
        self.assertEqual(options.getStore().journalMode, u'WAL')



# -*- test-case-name: axiom.test.test_queryutil -*-

import operator

from axiom.attributes import AND, OR

def contains(startAttribute,
             endAttribute,
             value):
    """
    Return an L{axiom.iaxiom.IComparison} (an object that can be
    passed as the 'comparison' argument to Store.query/.sum/.count)
    which will constrain a query against 2 attributes for ranges which
    contain the given argument.  The range is half-open.
    """
    return AND(
        startAttribute <= value,
        value < endAttribute)


def overlapping(startAttribute, # X
                endAttribute,   # Y
                startValue,     # A
                endValue,       # B
                ):
    """
    Return an L{axiom.iaxiom.IComparison} (an object that can be passed as the
    'comparison' argument to Store.query/.sum/.count) which will constrain a
    query against 2 attributes for ranges which overlap with the given
    arguments.

    For a database with Items of class O which represent values in this
    configuration::

              X                   Y
             (a)                 (b)
              |-------------------|
        (c)      (d)
         |--------|          (e)      (f)
                              |--------|

     (g) (h)
      |---|                            (i)    (j)
                                        |------|

     (k)                                   (l)
      |-------------------------------------|

             (a)                           (l)
              |-----------------------------|
        (c)                      (b)
         |------------------------|

        (c)  (a)
         |----|
                                 (b)       (l)
                                  |---------|

    The query::

        myStore.query(
            O,
            findOverlapping(O.X, O.Y,
                            a, b))

    Will return a generator of Items of class O which represent segments a-b,
    c-d, e-f, k-l, a-l, c-b, c-a and b-l, but NOT segments g-h or i-j.

    (NOTE: If you want to pass attributes of different classes for
    startAttribute and endAttribute, read the implementation of this method to
    discover the additional join clauses required.  This may be eliminated some
    day so for now, consider this method undefined over multiple classes.)

    In the database where this query is run, for an item N, all values of
    N.startAttribute must be less than N.endAttribute.

    startValue must be less than endValue.
    """
    assert startValue <= endValue

    return OR(
        AND(startAttribute >= startValue,
            startAttribute <= endValue),
        AND(endAttribute >= startValue,
            endAttribute <= endValue),
        AND(startAttribute <= startValue,
            endAttribute >= endValue)
        )

def _tupleCompare(tuple1, ineq, tuple2,
                 eq=lambda a,b: (a==b),
                 ander=AND,
                 orer=OR):
    """
    Compare two 'in-database tuples'.  Useful when sorting by a compound key
    and slicing into the middle of that query.
    """

    orholder = []
    for limit in range(len(tuple1)):
        eqconstraint = [
            eq(elem1, elem2) for elem1, elem2 in zip(tuple1, tuple2)[:limit]]
        ineqconstraint = ineq(tuple1[limit], tuple2[limit])
        orholder.append(ander(*(eqconstraint + [ineqconstraint])))
    return orer(*orholder)

def _tupleLessThan(tuple1, tuple2):
    return _tupleCompare(tuple1, operator.lt, tuple2)

def _tupleGreaterThan(tuple1, tuple2):
    return _tupleCompare(tuple1, operator.gt, tuple2)

class AttributeTuple(object):
    def __init__(self, *attributes):
        self.attributes = attributes

    def __iter__(self):
        return iter(self.attributes)

    def __eq__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return AND(*[
                myAttr == otherAttr
                for (myAttr, otherAttr)
                in zip(self, other)])

    def __ne__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return OR(*[
                myAttr != otherAttr
                for (myAttr, otherAttr)
                in zip(self, other)])

    def __gt__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return _tupleGreaterThan(tuple(iter(self)), other)

    def __lt__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return _tupleLessThan(tuple(iter(self)), other)

    def __ge__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return OR(self > other, self == other)

    def __le__(self, other):
        if not isinstance(other, (AttributeTuple, tuple, list)):
            return NotImplemented
        return OR(self < other, self == other)



# -*- test-case-name: axiom.test.test_upgrading -*-

"""
New version of L{axiom.test.oldapp} which upgrades swords by deleting them.
"""

from axiom.item import Item
from axiom.attributes import text
from axiom.upgrade import registerDeletionUpgrader

class Sword(Item):
    typeName = 'test_app_sword'
    schemaVersion = 2

    name = text()

registerDeletionUpgrader(Sword, 1, 2)



# -*- test-case-name: axiom.test.historic.test_parentHook2to3 -*-
"""
Generate a test stub for upgrading L{_SubSchedulerParentHook} from version 2 to
3, which removes the C{scheduledAt} attribute.
"""

from axiom.test.historic.stubloader import saveStub

from axiom.dependency import installOn
from axiom.scheduler import Scheduler, _SubSchedulerParentHook
from axiom.substore import SubStore

def createDatabase(store):
    scheduler = Scheduler(store=store)
    installOn(scheduler, store)
    installOn(
        _SubSchedulerParentHook(
            store=store, loginAccount=SubStore(store=store)),
        store)


if __name__ == '__main__':
    saveStub(createDatabase, 16800)



# Copyright 2008 Divmod, Inc.  See LICENSE for details.

"""
Test for the deletion of L{BatchManholePowerup}.
"""

from axiom.batch import BatchManholePowerup
from axiom.test.historic.stubloader import StubbedTest


class BatchManholePowerupTests(StubbedTest):
    def test_deletion(self):
        """
        The upgrade to schema version 2 deletes L{BatchManholePowerup}.
        """
        self.assertEqual(self.store.query(BatchManholePowerup).count(), 0)



# -*- test-case-name: axiom.test.test_attributes,axiom.test.test_reference -*-

import os

from decimal import Decimal

from epsilon import hotfix
hotfix.require('twisted', 'filepath_copyTo')

from zope.interface import implements

from twisted.python import filepath
from twisted.python.deprecate import deprecated
from twisted.python.components import registerAdapter
from twisted.python.versions import Version

from epsilon.extime import Time

from axiom.slotmachine import Attribute as inmemory

from axiom.errors import NoCrossStoreReferences, BrokenReference

from axiom.iaxiom import IComparison, IOrdering, IColumn, IQuery

_NEEDS_FETCH = object()         # token indicating that a value was not found

__metaclass__ = type


class _ComparisonOperatorMuxer:
    """
    Collapse comparison operations into calls to a single method with varying
    arguments.
    """
    def compare(self, other, op):
        """
        Override this in a subclass.
        """
        raise NotImplementedError()


    def __eq__(self, other):
        return self.compare(other, '=')


    def __ne__(self, other):
        return self.compare(other, '!=')


    def __gt__(self, other):
        return self.compare(other, '>')


    def __lt__(self, other):
        return self.compare(other, '<')


    def __ge__(self, other):
        return self.compare(other, '>=')


    def __le__(self, other):
        return self.compare(other, '<=')


def compare(left, right, op):
    # interim: maybe we want objects later?  right now strings should be fine
    if IColumn.providedBy(right):
        return TwoAttributeComparison(left, op, right)
    elif right is None:
        if op == '=':
            negate = False
        elif op == '!=':
            negate = True
        else:
            raise TypeError(
                "None/NULL does not work with %s comparison" % (op,))
        return NullComparison(left, negate)
    else:
        # convert to constant usable in the database
        return AttributeValueComparison(left, op, right)



class _MatchingOperationMuxer:
    """
    Collapse string matching operations into calls to a single method with
    varying arguments.
    """
    def _like(self, negate, firstOther, *others):
        others = (firstOther,) + others
        likeParts = []

        allValues = True
        for other in others:
            if IColumn.providedBy(other):
                likeParts.append(LikeColumn(other))
                allValues = False
            elif other is None:
                # LIKE NULL is a silly condition, but it's allowed.
                likeParts.append(LikeNull())
                allValues = False
            else:
                likeParts.append(LikeValue(other))

        if allValues:
            likeParts = [LikeValue(''.join(others))]

        return LikeComparison(self, negate, likeParts)


    def like(self, *others):
        return self._like(False, *others)


    def notLike(self, *others):
        return self._like(True, *others)


    def startswith(self, other):
        return self._like(False, other, '%')


    def endswith(self, other):
        return self._like(False, '%', other)



_ASC = 'ASC'
_DESC = 'DESC'

class _OrderingMixin:
    """
    Provide the C{ascending} and C{descending} attributes to specify sort
    direction.
    """
    def _asc(self):
        return SimpleOrdering(self, _ASC)

    def _desc(self):
        return SimpleOrdering(self, _DESC)

    desc = descending = property(_desc)
    asc = ascending = property(_asc)



class _ContainableMixin:
    def oneOf(self, seq, negate=False):
        """
        Choose items whose attributes are in a fixed set.

        X.oneOf([1, 2, 3])

        Implemented with the SQL 'in' statement.
        """
        return SequenceComparison(self, seq, negate)


    def notOneOf(self, seq):
        return self.oneOf(seq, negate=True)



class Comparable(_ContainableMixin, _ComparisonOperatorMuxer,
                 _MatchingOperationMuxer, _OrderingMixin):
    """
    Helper for a thing that can be compared like an SQLAttribute (or is in fact
    an SQLAttribute).  Requires that 'self' have 'type' (Item-subclass) and
    'columnName' (str) attributes, as well as an 'infilter' method in the
    spirit of SQLAttribute, documented below.
    """

    # XXX TODO: improve error reporting

    def compare(self, other, sqlop):
        return compare(self, other, sqlop)



class SimpleOrdering:
    """
    Currently this class is mostly internal.  More documentation will follow as
    its interface is finalized.
    """
    implements(IOrdering)

    # maybe this will be a useful public API, for the query something
    # something.

    isDescending = property(lambda self: self.direction == _DESC)
    isAscending = property(lambda self: self.direction == _ASC)

    def __init__(self, attribute, direction=''):
        self.attribute = attribute
        self.direction = direction


    def orderColumns(self):
        return [(self.attribute, self.direction)]


    def __repr__(self):
        return repr(self.attribute) + self.direction


    def __add__(self, other):
        if isinstance(other, SimpleOrdering):
            return CompoundOrdering([self, other])
        elif isinstance(other, (list, tuple)):
            return CompoundOrdering([self] + list(other))
        else:
            return NotImplemented


    def __radd__(self, other):
        if isinstance(other, SimpleOrdering):
            return CompoundOrdering([other, self])
        elif isinstance(other, (list, tuple)):
            return CompoundOrdering(list(other) + [self])
        else:
            return NotImplemented


class CompoundOrdering:
    """
    List of SimpleOrdering instances.
    """
    implements(IOrdering)

    def __init__(self, seq):
        self.simpleOrderings = list(seq)


    def __repr__(self):
        return self.__class__.__name__ + '(' + repr(self.simpleOrderings) + ')'


    def __add__(self, other):
        """
        Just thinking about what might be useful from the perspective of
        introspecting on query objects... don't document this *too* thoroughly
        yet.
        """
        if isinstance(other, CompoundOrdering):
            return CompoundOrdering(self.simpleOrderings + other.simpleOrderings)
        elif isinstance(other, SimpleOrdering):
            return CompoundOrdering(self.simpleOrderings + [other])
        elif isinstance(other, (list, tuple)):
            return CompoundOrdering(self.simpleOrderings + list(other))
        else:
            return NotImplemented


    def __radd__(self, other):
        """
        Just thinking about what might be useful from the perspective of
        introspecting on query objects... don't document this *too* thoroughly
        yet.
        """
        if isinstance(other, CompoundOrdering):
            return CompoundOrdering(other.simpleOrderings + self.simpleOrderings)
        elif isinstance(other, SimpleOrdering):
            return CompoundOrdering([other] + self.simpleOrderings)
        elif isinstance(other, (list, tuple)):
            return CompoundOrdering(list(other) + self.simpleOrderings)
        else:
            return NotImplemented


    def orderColumns(self):
        x = []
        for o in self.simpleOrderings:
            x.extend(o.orderColumns())
        return x



class UnspecifiedOrdering:
    implements(IOrdering)

    def __init__(self, null):
        pass

    def __add__(self, other):
        return IOrdering(other, NotImplemented)

    __radd__ = __add__


    def orderColumns(self):
        return []


registerAdapter(CompoundOrdering, list, IOrdering)
registerAdapter(CompoundOrdering, tuple, IOrdering)
registerAdapter(UnspecifiedOrdering, type(None), IOrdering)
registerAdapter(SimpleOrdering, Comparable, IOrdering)

def compoundIndex(*columns):
    for column in columns:
        column.compoundIndexes.append(columns)

class SQLAttribute(inmemory, Comparable):
    """
    Abstract superclass of all attributes.

    _Not_ an attribute itself.

    @ivar indexed: A C{bool} indicating whether this attribute will be indexed
    in the database.

    @ivar default: The value used for this attribute, if no value is specified.
    """
    implements(IColumn)

    sqltype = None

    def __init__(self, doc='', indexed=False, default=None, allowNone=True, defaultFactory=None):
        inmemory.__init__(self, doc)
        self.indexed = indexed
        self.compoundIndexes = []
        self.allowNone = allowNone
        self.default = default
        self.defaultFactory = defaultFactory
        if default is not None and defaultFactory is not None:
            raise ValueError("You may specify only one of default "
                             "or defaultFactory, not both")

    def computeDefault(self):
        if self.defaultFactory is not None:
            return self.defaultFactory()
        return self.default


    def reprFor(self, oself):
        return repr(self.__get__(oself))


    def getShortColumnName(self, store):
        return store.getShortColumnName(self)


    def getColumnName(self, store):
        return store.getColumnName(self)


    def prepareInsert(self, oself, store):
        """
        Override this method to do something to an item to prepare for its
        insertion into a database.
        """

    def coercer(self, value):
        """
        must return a value equivalent to the data being passed in for it to be
        considered valid for a value of this attribute.  for example, 'int' or
        'str'.
        """

        raise NotImplementedError()


    def infilter(self, pyval, oself, store):
        """
        used to convert a Python value to something that lives in the database;
        so called because it is called when objects go in to the database.  It
        takes a Python value and returns an SQL value.
        """
        raise NotImplementedError()

    def outfilter(self, dbval, oself):
        """
        used to convert an SQL value to something that lives in memory; so
        called because it is called when objects come out of the database.  It
        takes an SQL value and returns a Python value.
        """
        return dbval

    # requiredSlots must be called before it's run

    prefix = "_axiom_memory_"
    dbprefix = "_axiom_store_"

    def requiredSlots(self, modname, classname, attrname):
        self.modname = modname
        self.classname = classname
        self.attrname = attrname
        self.underlying = self.prefix + attrname
        self.dbunderlying = self.dbprefix + attrname
        yield self.underlying
        yield self.dbunderlying


    def fullyQualifiedName(self):
        return '.'.join([self.modname,
                         self.classname,
                         self.attrname])

    def __repr__(self):
        return '<%s %s>' % ( self.__class__.__name__, self.fullyQualifiedName())

    def type():
        def get(self):
            if self._type is None:
                from twisted.python.reflect import namedAny
                self._type = namedAny(self.modname+'.'+self.classname)
            return self._type
        return get,
    _type = None
    type = property(*type())

    def __get__(self, oself, cls=None):
        if cls is not None and oself is None:
            if self._type is not None:
                assert self._type == cls
            else:
                self._type = cls
            return self

        pyval = getattr(oself, self.underlying, _NEEDS_FETCH)
        if pyval is _NEEDS_FETCH:
            dbval = getattr(oself, self.dbunderlying, _NEEDS_FETCH)
            if dbval is _NEEDS_FETCH:
                # here is what *is* happening here:

                # SQL attributes are always loaded when an Item is created by
                # loading from the database, either via a query, a getItemByID
                # or an attribute access.  If an attribute is left un-set, that
                # means that the item it is on was just created, and we fill in
                # the default value.

                # Here is what *should be*, but *is not* happening here:

                # this condition ought to indicate that a value may exist in
                # the database, but it is not currently available in memory.
                # It would then query the database immediately, loading all
                # SQL-resident attributes related to this item to minimize the
                # number of queries run (e.g. rather than one per attribute)

                # this is a more desireable condition because it means that you
                # can create items "for free", so doing, for example,
                # self.bar.storeID is a much cheaper operation than doing
                # self.bar.baz.  This particular idiom is frequently used in
                # queries and so speeding it up to avoid having to do a
                # database hit unless you actually need an item's attributes
                # would be worthwhile.

                return self.default
            pyval = self.outfilter(dbval, oself)
            # An upgrader may have changed the value of this attribute.  If so,
            # return the new value, not the old one.
            if dbval != getattr(oself, self.dbunderlying):
                return self.__get__(oself, cls)
            # cache python value
            setattr(oself, self.underlying, pyval)
        return pyval

    def loaded(self, oself, dbval):
        """
        This method is invoked when the item is loaded from the database, and
        when a transaction is reverted which restores this attribute's value.

        @param oself: an instance of an item which has this attribute.

        @param dbval: the underlying database value which was retrieved.
        """
        setattr(oself, self.dbunderlying, dbval)
        delattr(oself, self.underlying) # member_descriptors don't raise
                                        # attribute errors; what gives?  good
                                        # for us, I guess.

    def _convertPyval(self, oself, pyval):
        """
        Convert a Python value to a value suitable for inserting into the
        database.

        @param oself: The object on which this descriptor is an attribute.
        @param pyval: The value to be converted.
        @return: A value legal for this column in the database.
        """
        # convert to dbval later, I guess?
        if pyval is None and not self.allowNone:
            raise TypeError("attribute [%s.%s = %s()] must not be None" % (
                    self.classname, self.attrname, self.__class__.__name__))

        return self.infilter(pyval, oself, oself.store)

    def __set__(self, oself, pyval):
        st = oself.store

        dbval = self._convertPyval(oself, pyval)
        oself.__dirty__[self.attrname] = self, dbval
        oself.touch()
        setattr(oself, self.underlying, pyval)
        setattr(oself, self.dbunderlying, dbval)
        if st is not None and st.autocommit:
            st._rejectChanges += 1
            try:
                oself.checkpoint()
            finally:
                st._rejectChanges -= 1


class TwoAttributeComparison:
    implements(IComparison)
    def __init__(self, leftAttribute, operationString, rightAttribute):
        self.leftAttribute = leftAttribute
        self.operationString = operationString
        self.rightAttribute = rightAttribute

    def getQuery(self, store):
        sql = ('(%s %s %s)' % (self.leftAttribute.getColumnName(store),
                               self.operationString,
                               self.rightAttribute.getColumnName(store)) )
        return sql

    def getInvolvedTables(self):
        tables = [self.leftAttribute.type]
        if self.leftAttribute.type is not self.rightAttribute.type:
            tables.append(self.rightAttribute.type)
        return tables


    def getArgs(self, store):
        return []


    def __repr__(self):
        return ' '.join((self.leftAttribute.fullyQualifiedName(),
                         self.operationString,
                         self.rightAttribute.fullyQualifiedName()))


class AttributeValueComparison:
    implements(IComparison)
    def __init__(self, attribute, operationString, value):
        self.attribute = attribute
        self.operationString = operationString
        self.value = value

    def getQuery(self, store):
        return ('(%s %s ?)' % (self.attribute.getColumnName(store),
                               self.operationString))

    def getArgs(self, store):
        return [self.attribute.infilter(self.value, None, store)]

    def getInvolvedTables(self):
        return [self.attribute.type]

    def __repr__(self):
        return ' '.join((self.attribute.fullyQualifiedName(),
                         self.operationString,
                         repr(self.value)))

class NullComparison:
    implements(IComparison)
    def __init__(self, attribute, negate=False):
        self.attribute = attribute
        self.negate = negate

    def getQuery(self, store):
        if self.negate:
            op = 'NOT'
        else:
            op = 'IS'
        return ('(%s %s NULL)' % (self.attribute.getColumnName(store),
                                  op))

    def getArgs(self, store):
        return []

    def getInvolvedTables(self):
        return [self.attribute.type]

class LikeFragment:
    def getLikeArgs(self):
        return []

    def getLikeQuery(self, st):
        raise NotImplementedError()

    def getLikeTables(self):
        return []

class LikeNull(LikeFragment):
    def getLikeQuery(self, st):
        return "NULL"

class LikeValue(LikeFragment):
    def __init__(self, value):
        self.value = value

    def getLikeQuery(self, st):
        return "?"

    def getLikeArgs(self):
        return [self.value]

class LikeColumn(LikeFragment):
    def __init__(self, attribute):
        self.attribute = attribute

    def getLikeQuery(self, st):
        return self.attribute.getColumnName(st)

    def getLikeTables(self):
        return [self.attribute.type]


class LikeComparison:
    implements(IComparison)
    # Not AggregateComparison or AttributeValueComparison because there is a
    # different, optimized syntax for 'or'.  WTF is wrong with you, SQL??

    def __init__(self, attribute, negate, likeParts):
        self.negate = negate
        self.attribute = attribute
        self.likeParts = likeParts

    def getInvolvedTables(self):
        tables = [self.attribute.type]
        for lf in self.likeParts:
            tables.extend([
                    t for t in lf.getLikeTables() if t not in tables])
        return tables

    def getQuery(self, store):
        if self.negate:
            op = 'NOT LIKE'
        else:
            op = 'LIKE'
        sqlParts = [lf.getLikeQuery(store) for lf in self.likeParts]
        sql = '(%s %s (%s))' % (self.attribute.getColumnName(store),
                                op, ' || '.join(sqlParts))
        return sql

    def getArgs(self, store):
        l = []
        for lf in self.likeParts:
            for pyval in lf.getLikeArgs():
                l.append(
                    self.attribute.infilter(
                        pyval, None, store))
        return l



class AggregateComparison:
    """
    Abstract base class for compound comparisons that aggregate other
    comparisons - currently only used for AND and OR comparisons.
    """

    implements(IComparison)
    operator = None

    def __init__(self, *conditions):
        self.conditions = conditions
        if self.operator is None:
            raise NotImplementedError, ('%s cannot be used; you want AND or OR.'
                                        % self.__class__.__name__)
        if not conditions:
            raise ValueError, ('%s condition requires at least one argument'
                               % self.operator)

    def getQuery(self, store):
        oper = ' %s ' % self.operator
        return '(%s)' % oper.join(
            [condition.getQuery(store) for condition in self.conditions])

    def getArgs(self, store):
        args = []
        for cond in self.conditions:
            args += cond.getArgs(store)
        return args

    def getInvolvedTables(self):
        tables = []
        for cond in self.conditions:
            tables.extend([
                    t for t in cond.getInvolvedTables() if t not in tables])
        return tables

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__,
                           ', '.join(map(repr, self.conditions)))



class SequenceComparison:
    implements(IComparison)

    def __init__(self, attribute, container, negate):
        self.attribute = attribute
        self.container = container
        self.negate = negate

        if IColumn.providedBy(container):
            self.containerClause = self._columnContainer
            self.getArgs = self._columnArgs
        elif IQuery.providedBy(container):
            self.containerClause = self._queryContainer
            self.getArgs = self._queryArgs
        else:
            self.containerClause = self._sequenceContainer
            self.getArgs = self._sequenceArgs


    def _columnContainer(self, store):
        """
        Return the fully qualified name of the column being examined so as
        to push all of the containment testing into the database.
        """
        return self.container.getColumnName(store)


    def _columnArgs(self, store):
        """
        The IColumn form of this has no arguments, just a column name
        specified in the SQL, specified by _columnContainer.
        """
        return []


    _subselectSQL = None
    _subselectArgs = None
    def _queryContainer(self, store):
        """
        Generate and cache the subselect SQL and its arguments.  Return the
        subselect SQL.
        """
        if self._subselectSQL is None:
            sql, args = self.container._sqlAndArgs('SELECT',
                                                   self.container._queryTarget)
            self._subselectSQL, self._subselectArgs = sql, args
        return self._subselectSQL


    def _queryArgs(self, store):
        """
        Make sure subselect arguments have been generated and then return
        them.
        """
        self._queryContainer(store)
        return self._subselectArgs


    _sequence = None
    def _sequenceContainer(self, store):
        """
        Smash whatever we got into a list and save the result in case we are
        executed multiple times.  This keeps us from tripping up over
        generators and the like.
        """
        if self._sequence is None:
            self._sequence = list(self.container)
            self._clause = ', '.join(['?'] * len(self._sequence))
        return self._clause


    def _sequenceArgs(self, store):
        """
        Filter each element of the data using the attribute type being
        tested for containment and hand back the resulting list.
        """
        self._sequenceContainer(store) # Force _sequence to be valid
        return [self.attribute.infilter(pyval, None, store) for pyval in self._sequence]


    # IComparison - getArgs is assigned as an instance attribute
    def getQuery(self, store):
        return '%s %sIN (%s)' % (
            self.attribute.getColumnName(store),
            self.negate and 'NOT ' or '',
            self.containerClause(store))


    def getInvolvedTables(self):
        return [self.attribute.type]



class AND(AggregateComparison):
    """
    Combine 2 L{IComparison}s such that this is true when both are true.
    """
    operator = 'AND'

class OR(AggregateComparison):
    """
    Combine 2 L{IComparison}s such that this is true when either is true.
    """
    operator = 'OR'


class TableOrderComparisonWrapper(object):
    """
    Wrap any other L{IComparison} and override its L{getInvolvedTables} method
    to specify the same tables but in an explicitly specified order.
    """
    implements(IComparison)

    tables = None
    comparison = None

    def __init__(self, tables, comparison):
        assert set(tables) == set(comparison.getInvolvedTables())

        self.tables = tables
        self.comparison = comparison


    def getInvolvedTables(self):
        return self.tables


    def getQuery(self, store):
        return self.comparison.getQuery(store)


    def getArgs(self, store):
        return self.comparison.getArgs(store)



class boolean(SQLAttribute):
    sqltype = 'BOOLEAN'

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        if pyval is True:
            return 1
        elif pyval is False:
            return 0
        else:
            raise TypeError("attribute [%s.%s = boolean()] must be True or False; not %r" %
                            (self.classname, self.attrname, type(pyval).__name__,))

    def outfilter(self, dbval, oself):
        if dbval == 1:
            return True
        elif dbval == 0:
            return False
        elif self.allowNone and dbval is None:
            return None
        else:
            raise ValueError(
                "attribute [%s.%s = boolean()] "
                "must have a database value of 1 or 0; not %r" %
                (self.classname, self.attrname, dbval))



LARGEST_POSITIVE = (2 ** 63)-1
LARGEST_NEGATIVE = -(2 ** 63)

class ConstraintError(TypeError):
    """A type constraint was violated.
    """

    def __init__(self,
                 attributeObj,
                 requiredTypes,
                 providedValue):
        self.attributeObj = attributeObj
        self.requiredTypes = requiredTypes
        self.providedValue = providedValue
        TypeError.__init__(self,
                           "attribute [%s.%s = %s()] must be "
                           "(%s); not %r" %
                           (attributeObj.classname,
                            attributeObj.attrname,
                            attributeObj.__class__.__name__,
                            requiredTypes,
                            type(providedValue).__name__))



def requireType(attributeObj, value, typerepr, *types):
    if not isinstance(value, types):
        raise ConstraintError(attributeObj,
                              typerepr,
                              value)



inttyperepr = "integer between %r and %r" % (LARGEST_NEGATIVE, LARGEST_POSITIVE)

class integer(SQLAttribute):
    sqltype = 'INTEGER'
    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        requireType(self, pyval, inttyperepr, int, long)
        if not LARGEST_NEGATIVE <= pyval <= LARGEST_POSITIVE:
            raise ConstraintError(
                self, inttyperepr, pyval)
        return pyval



class bytes(SQLAttribute):
    """
    Attribute representing a sequence of bytes; this is represented in memory
    as a Python 'str'.
    """

    sqltype = 'BLOB'

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        if isinstance(pyval, unicode):
            raise ConstraintError(self, "str or other byte buffer", pyval)
        return buffer(pyval)


    def outfilter(self, dbval, oself):
        if dbval is None:
            return None
        return str(dbval)


    @deprecated(Version("Axiom", 0, 7, 5))
    def like(self, *others):
        return super(SQLAttribute, self).like(*others)


    @deprecated(Version("Axiom", 0, 7, 5))
    def notLike(self, *others):
        return super(SQLAttribute, self).notLike(*others)


    @deprecated(Version("Axiom", 0, 7, 5))
    def startswith(self, other):
        return super(SQLAttribute, self).startswith(other)


    @deprecated(Version("Axiom", 0, 7, 5))
    def endswith(self, other):
        return super(SQLAttribute, self).endswith(other)



class InvalidPathError(ValueError):
    """
    A path that could not be used with the database was attempted to be used
    with the database.
    """

class text(SQLAttribute):
    """
    Attribute representing a sequence of characters; this is represented in
    memory as a Python 'unicode'.
    """

    def __init__(self, caseSensitive=False, **kw):
        SQLAttribute.__init__(self, **kw)
        if caseSensitive:
            self.sqltype = 'TEXT'
        else:
            self.sqltype = 'TEXT COLLATE NOCASE'
        self.caseSensitive = caseSensitive

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        if not isinstance(pyval, unicode) or u'\0' in pyval:
            raise ConstraintError(
                self, "unicode string without NULL bytes", pyval)
        return pyval

    def outfilter(self, dbval, oself):
        return dbval



class textlist(text):
    delimiter = u'\u001f'

    # Once upon a time, textlist encoded the list in such a way that caused []
    # to be indistinguishable from [u'']. This value is now used as a
    # placeholder at the head of the list, to avoid this problem in a way that
    # is almost completely backwards-compatible with older databases.
    guard = u'\u0002'

    def outfilter(self, dbval, oself):
        unicodeString = super(textlist, self).outfilter(dbval, oself)
        if unicodeString is None:
            return None
        val = unicodeString.split(self.delimiter)
        if val[:1] == [self.guard]:
            del val[:1]
        return val

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        for innerVal in pyval:
            assert self.delimiter not in innerVal and self.guard not in innerVal
        result = self.delimiter.join([self.guard] + list(pyval))
        return super(textlist, self).infilter(result, oself, store)

class path(text):
    """
    Attribute representing a pathname in the filesystem.  If 'relative=True',
    the default, the representative pathname object must be somewhere inside
    the store, and will migrate with the store.

    I expect L{twisted.python.filepath.FilePath} or compatible objects as my
    values.
    """

    def __init__(self, relative=True, **kw):
        text.__init__(self, **kw)
        self.relative = True

    def prepareInsert(self, oself, store):
        """
        Prepare for insertion into the database by making the dbunderlying
        attribute of the item a relative pathname with respect to the store
        rather than an absolute pathname.
        """
        if self.relative:
            fspath = self.__get__(oself)
            oself.__dirty__[self.attrname] = self, self.infilter(fspath, oself, store)

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        mypath = unicode(pyval.path)
        if store is None:
            store = oself.store
        if store is None:
            return None
        if self.relative:
            # XXX add some more filepath APIs to make this kind of checking easier.
            storepath = os.path.normpath(store.filesdir.path)
            mysegs = mypath.split(os.sep)
            storesegs = storepath.split(os.sep)
            if len(mysegs) <= len(storesegs) or mysegs[:len(storesegs)] != storesegs:
                raise InvalidPathError('%s not in %s' % (mypath, storepath))
            # In the database we use '/' to separate paths for portability.
            # This databaes could have relative paths created on Windows, then
            # be moved to Linux for deployment, and what *was* the native
            # os.sep (backslash) will not be friendly to Linux's filesystem.
            # However, this is only for relative paths, since absolute or UNC
            # pathnames on a Windows system are inherently unportable and it's
            # not reasonable to calculate relative paths outside the store.
            p = '/'.join(mysegs[len(storesegs):])
        else:
            p = mypath          # we already know it's absolute, it came from a
                                # filepath.
        return super(path, self).infilter(p, oself, store)

    def outfilter(self, dbval, oself):
        if dbval is None:
            return None
        if self.relative:
            fp = oself.store.filesdir
            for segment in dbval.split('/'):
                fp = fp.child(segment)
        else:
            fp = filepath.FilePath(dbval)
        return fp


MICRO = 1000000.

class timestamp(integer):
    """
    An in-database representation of date and time.

    To make formatting as easy as possible, this is represented in Python as an
    instance of L{epsilon.extime.Time}; see its documentation for more details.
    """
    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        return integer.infilter(self,
                                int(pyval.asPOSIXTimestamp() * MICRO), oself,
                                store)

    def outfilter(self, dbval, oself):
        if dbval is None:
            return None
        return Time.fromPOSIXTimestamp(dbval / MICRO)

_cascadingDeletes = {}
_disallows = {}

class reference(integer):
    NULLIFY = object()
    DISALLOW = object()
    CASCADE = object()

    def __init__(self, doc='', indexed=True, allowNone=True, reftype=None,
                 whenDeleted=NULLIFY):
        integer.__init__(self, doc, indexed, None, allowNone)
        assert whenDeleted in (reference.NULLIFY,
                               reference.CASCADE,
                               reference.DISALLOW),(
            "whenDeleted must be one of: "
            "reference.NULLIFY, reference.CASCADE, reference.DISALLOW")
        self.reftype = reftype
        self.whenDeleted = whenDeleted
        if whenDeleted is reference.CASCADE:
            # Note; this list is technically in a slightly inconsistent state
            # as things are being built.
            _cascadingDeletes.setdefault(reftype, []).append(self)
        if whenDeleted is reference.DISALLOW:
            _disallows.setdefault(reftype, []).append(self)

    def reprFor(self, oself):
        obj = getattr(oself, self.underlying, None)
        if obj is not None:
            if obj.storeID is not None:
                return 'reference(%d)' % (obj.storeID,)
            else:
                return 'reference(unstored@%d)' % (id(obj),)
        sid = getattr(oself, self.dbunderlying, None)
        if sid is None:
            return 'None'
        return 'reference(%d)' % (sid,)


    def __get__(self, oself, cls=None):
        """
        Override L{integer.__get__} to verify that the value to be returned is
        currently a valid item in the same store, and to make sure that legacy
        items are upgraded if they happen to have been cached.
        """
        rv = super(reference, self).__get__(oself, cls)
        if rv is self:
            # If it's an attr lookup on the class, just do that.
            return self
        if rv is None:
            return rv
        if not rv._currentlyValidAsReferentFor(oself.store):
            # Make sure it's currently valid, i.e. it's not going to be deleted
            # this transaction or it hasn't been deleted.

            # XXX TODO: drop cached in-memory referent if it's been deleted /
            # no longer valid.
            assert self.whenDeleted is reference.NULLIFY, (
                "not sure what to do if not...")
            return None
        # If the current cached value is a legacy item, we discard it in order
        # to force another fetch from the database. However, if *this item* is
        # also a legacy item, then the item referred to may have been created
        # in an upgrader and not have been stored yet, so we shouldn't discard
        # it.
        if rv.__legacy__ and not oself.__legacy__:
            delattr(oself, self.underlying)
            return super(reference, self).__get__(oself, cls)
        return rv

    def prepareInsert(self, oself, store):
        oitem = super(reference, self).__get__(oself) # bypass NULLIFY
        if oitem is not None and oitem.store is not store:
            raise NoCrossStoreReferences(
                "Trying to insert item: %r into store: %r, "
                "but it has a reference to other item: .%s=%r "
                "in another store: %r" % (
                    oself, store,
                    self.attrname, oitem,
                    oitem.store))

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        if oself is None:
            return pyval.storeID
        if oself.store is None:
            return pyval.storeID
        if oself.store != pyval.store:
            raise NoCrossStoreReferences(
                "You can't establish references to items in other stores.")

        return integer.infilter(self, pyval.storeID, oself, store)

    def outfilter(self, dbval, oself):
        if dbval is None:
            return None

        referee = oself.store.getItemByID(dbval, default=None, autoUpgrade=not oself.__legacy__)
        if referee is None and self.whenDeleted is not reference.NULLIFY:

            # If referee merely changed to another valid referent,
            # SQLAttribute.__get__ will notice that what we returned is
            # inconsistent and try again.  However, it doesn't know about the
            # BrokenReference that is raised if the old referee is no longer a
            # valid referent.  Check to see if the dbunderlying is still the
            # same as the dbval passed in.  If it's different, we should try to
            # load the value again.  Only if it is unchanged will we raise the
            # BrokenReference.  It would be better if all of this
            # change-detection logic were in one place, but I can't figure out
            # how to do that. -exarkun
            if dbval != getattr(oself, self.dbunderlying):
                return self.__get__(oself, None)

            raise BrokenReference('Reference to storeID %r is broken' % (dbval,))
        return referee

class ieee754_double(SQLAttribute):
    """
    From the SQLite documentation::

        Each value stored in an SQLite database (or manipulated by the
        database engine) has one of the following storage classes: (...)
        REAL. The value is a floating point value, stored as an 8-byte IEEE
        floating point number.

    This attribute type implements IEEE754 double-precision binary
    floating-point storage.  Some people call this 'float', and think it is
    somehow related to numbers.  This assumption can be misleading when working
    with certain types of data.

    This attribute name has an unweildy name on purpose.  You should be aware
    of the caveats related to binary floating point math before using this
    type.  It is particularly ill-advised to use it to store values
    representing large amounts of currency as rounding errors may be
    significant enough to introduce accounting discrepancies.

    Certain edge-cases are not handled properly.  For example, INF and NAN are
    considered by SQLite to be equal to everything, rather than the Python
    interpretation where INF is equal only to itself and greater than
    everything, and NAN is equal to nothing, not even itself.
    """

    sqltype = 'REAL'

    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        requireType(self, pyval, 'float', float)
        return pyval

    def outfilter(self, dbval, oself):
        return dbval



class AbstractFixedPointDecimal(integer):
    """
    Attribute representing a number with a specified number of decimal
    places.

    This is stored in SQLite as a binary integer multiplied by M{10**N}
    where C{N} is the number of decimal places required by Python. 
    Therefore, in-database multiplication, division, or queries which
    compare to integers or fixedpointdecimals with a different number of
    decimal places, will not work.  Also, you cannot store, or sum to, fixed
    point decimals greater than M{(2**63)/(10**N)}.

    While L{ieee754_double} is handy for representing various floating-point
    numbers, such as scientific measurements, this class (and the associated
    Python decimal class) is more appropriate for arithmetic on sums of money.

    For more information on Python's U{Decimal
    class<http://www.python.org/doc/current/lib/module-decimal.html>} and on
    general U{computerized Decimal math in
    general<http://www2.hursley.ibm.com/decimal/decarith.html>}.

    This is currently a private helper superclass because we cannot store
    additional metadata about column types; maybe we should fix that.

    @cvar decimalPlaces: the number of points of decimal precision allowed by
    the storage and retrieval of this class.  *Points beyond this number
    will be silently truncated to values passed into the database*, so be
    sure to select a value appropriate to your application!
    """

    def __init__(self, **kw):
        integer.__init__(self, **kw)


    def infilter(self, pyval, oself, store):
        if pyval is None:
            return None
        if isinstance(pyval, (int, long)):
            pyval = Decimal(pyval)
        if isinstance(pyval, Decimal):
            # Python < 2.5.2 compatibility:
            # Use to_integral instead of to_integral_value.
            dbval = int((pyval * 10**self.decimalPlaces).to_integral())
            return super(AbstractFixedPointDecimal, self).infilter(
                dbval, oself, store)
        else:
            raise TypeError(
                "attribute [%s.%s = AbstractFixedPointDecimal(...)] must be "
                "Decimal instance; not %r" % (
                self.classname, self.attrname, type(pyval).__name__))


    def outfilter(self, dbval, oself):
        if dbval is None:
            return None
        return Decimal(dbval) / 10**self.decimalPlaces


    def compare(self, other, sqlop):
        if isinstance(other, Comparable):
            if isinstance(other, AbstractFixedPointDecimal):
                if other.decimalPlaces == self.decimalPlaces:
                    # fall through to default behavior at bottom
                    pass
                else:
                    raise TypeError(
                        "Can't compare Decimals of varying precisions: "
                        "(%s.%s %s %s.%s)" % (
                            self.classname, self.attrname,
                            sqlop,
                            other.classname, other.attrname
                            ))
            else:
                raise TypeError(
                    "Can't compare Decimals to other things: "
                    "(%s.%s %s %s.%s)" % (
                        self.classname, self.attrname,
                        sqlop,
                        other.classname, other.attrname
                        ))
        return super(AbstractFixedPointDecimal, self).compare(other, sqlop)

class point1decimal(AbstractFixedPointDecimal):
    decimalPlaces = 1
class point2decimal(AbstractFixedPointDecimal):
    decimalPlaces = 2
class point3decimal(AbstractFixedPointDecimal):
    decimalPlaces = 3
class point4decimal(AbstractFixedPointDecimal):
    decimalPlaces = 4
class point5decimal(AbstractFixedPointDecimal):
    decimalPlaces = 5
class point6decimal(AbstractFixedPointDecimal):
    decimalPlaces = 6
class point7decimal(AbstractFixedPointDecimal):
    decimalPlaces = 7
class point8decimal(AbstractFixedPointDecimal):
    decimalPlaces = 8
class point9decimal(AbstractFixedPointDecimal):
    decimalPlaces = 9
class point10decimal(AbstractFixedPointDecimal):
    decimalPlaces = 10

class money(point4decimal):
    """
    I am a 4-point precision fixed-point decimal number column type; suggested
    for representing a quantity of money.

    (This does not, however, include features such as currency.)
    """



# Copright 2008 Divmod, Inc.  See LICENSE file for details.
# -*- test-case-name: axiom.test.test_userbase -*-

"""
The L{axiom.userbase} module implements various interfaces from L{twisted.cred}
to allow an Axiom database to serve as an integration point for Twisted
services that do authentication.

While not strictly required, one part of this implementation is the idiom that
Axiom (by default) partitions its user database into a separate data-store for
each users.

This has several advantages:

  - Each user's account can be quickly and independently added to or removed
    from the system; inactive accounts can be quickly moved to archival
    storage.

  - User accounts may be migrated between servers relatively easily.

  - Database queries that deal with a single user's data are completely
    partitioned; even naive and inefficient queries can still be run quickly as
    long as users do not individually have a lot of data in a particular table.

For truly multi-user applications, this partitioning is incomplete without an
abstract facility for exchanging data between different users of the same
application.  This module does not implement such a facility, as it is left to
higher-level mechanisms such as Mantissa's messaging system in
L{xmantissa.messaging}.  However, this module works standalone as well; just be
aware that a user's database contains only their own data.
"""

import warnings

from zope.interface import implements, Interface

from twisted.cred.portal import IRealm
from twisted.cred.credentials import IUsernamePassword, IUsernameHashedPassword
from twisted.cred.checkers import ICredentialsChecker, ANONYMOUS
from twisted.python import log
from twisted.internet.defer import succeed, fail

from axiom.store import Store
from axiom.item import Item
from axiom.substore import SubStore
from axiom.attributes import text, integer, reference, boolean, AND, OR
from axiom.errors import (
    BadCredentials, NoSuchUser, DuplicateUser, MissingDomainPart)
from axiom.scheduler import IScheduler
from axiom import upgrade, iaxiom

ANY_PROTOCOL = u'*'

def dflip(x):
    warnings.warn("Don't use dflip no more", stacklevel=2)
    return x


class AllNamesConflict(Exception):
    """
    When inserting a SubStore into a site store, no names were found which were
    not already associated with an account.

    This prevents the SubStore from being inserted at all.  No files are moved
    and the site database is not modified.
    """


class DatabaseDirectoryConflict(Exception):
    """
    When inserting a SubStore into a site store, the selected ultimate location
    for the SubStore's Axiom database directory already existed.

    This prevents the SubStore from being inserted at all.  No files are moved
    and the site database is not modified.
    """



class IPreauthCredentials(Interface):
    """
    Deprecated.  Don't use this.  If you wrote a checker which can check this
    interface, make it check one of the interfaces L{Preauthenticated}
    implements, instead.
    """



class Preauthenticated(object):
    """
    A credentials object of multiple types which has already been authenticated
    somehow.

    Credentials interfaces methods are implemented to behave as if the correct
    credentials had been supplied.
    """
    implements(IUsernamePassword, IUsernameHashedPassword)

    def __init__(self, username):
        self.username = username


    def checkPassword(self, password):
        """
        The password checks out.
        """
        return True


    def __repr__(self):
        return '<Preauthenticated: %s>' % (self.username,)



class LoginMethod(Item):
    typeName = 'login_method'
    schemaVersion = 2

    localpart = text(doc="""
    A local-part of my user's identifier.
    """, indexed=True, allowNone=False)

    domain = text(doc="""
    The domain part of my user's identifier. [XXX See TODO below]
    May be None (generally for "system" users).
    """, indexed=True)

    internal = boolean(doc="""
    Flag indicating whether this is a method maintained by this server, or if
    it represents an external contact mechanism (such as a third-party email
    provider)
    """, allowNone=False)

    protocol = text(indexed=True, allowNone=False)
    account = reference(doc="""
    A reference to the LoginAccount for which this is a login method.
    """, allowNone=False)

    verified = boolean(indexed=True, allowNone=False)

def upgradeLoginMethod1To2(old):
    return old.upgradeVersion(
            'login_method', 1, 2,
            localpart=old.localpart,
            domain=old.domain,
            internal=old.internal,
            protocol=old.protocol,
            account=old.account,
            verified=old.verified)

upgrade.registerUpgrader(upgradeLoginMethod1To2, 'login_method', 1, 2)

class LoginAccount(Item):
    """
    I am an entry in a LoginBase.

    @ivar avatars: An Item which is adaptable to various cred client
    interfaces.  Plural because it represents a collection of potentially
    disparate implementors, such as an IResource for web access and an IContact
    for SIP access.

    @ivar disabled: This account has been disabled.  It is still
    database-resident but the user should not be allowed to log in.

    """
    typeName = 'login'
    schemaVersion = 2

    # DEPRECATED: Do not use directly; see setPassword etc.
    password = text()

    avatars = reference()       # reference to a thing which can be adapted to
                                # implementations for application-level
                                # protocols.  In general this is a reference to
                                # a SubStore because this is optimized for
                                # applications where per-user data is a
                                # substantial portion of the cost.
    disabled = integer()


    def __conform__(self, interface):
        """
        For convenience, forward adaptation to my 'avatars' attribute.
        """
        ifa = interface(self.avatars, None)
        return ifa


    def migrateDown(self):
        """
        Assuming that self.avatars is a SubStore which should contain *only*
        the LoginAccount for the user I represent, remove all LoginAccounts and
        LoginMethods from that store and copy all methods from the site store
        down into it.
        """
        ss = self.avatars.open()
        def _():
            oldAccounts = ss.query(LoginAccount)
            oldMethods = ss.query(LoginMethod)
            for x in list(oldAccounts) + list(oldMethods):
                x.deleteFromStore()
            self.cloneInto(ss, ss)
            IScheduler(ss).migrateDown()
        ss.transact(_)


    def migrateUp(self):
        """
        Copy this LoginAccount and all associated LoginMethods from my store
        (which is assumed to be a SubStore, most likely a user store) into the
        site store which contains it.
        """
        siteStore = self.store.parent
        def _():
            # No convenience method for the following because needing to do it is
            # *rare*.  It *should* be ugly; 99% of the time if you need to do this
            # you're making a mistake. -glyph
            siteStoreSubRef = siteStore.getItemByID(self.store.idInParent)
            self.cloneInto(siteStore, siteStoreSubRef)
            IScheduler(self.store).migrateUp()
        siteStore.transact(_)


    def cloneInto(self, newStore, avatars):
        """
        Create a copy of this LoginAccount and all associated LoginMethods in a different Store.

        Return the copied LoginAccount.
        """
        la = LoginAccount(store=newStore,
                          password=self.password,
                          avatars=avatars,
                          disabled=self.disabled)
        for siteMethod in self.store.query(LoginMethod,
                                           LoginMethod.account == self):
            LoginMethod(store=newStore,
                        localpart=siteMethod.localpart,
                        domain=siteMethod.domain,
                        internal=siteMethod.internal,
                        protocol=siteMethod.protocol,
                        verified=siteMethod.verified,
                        account=la)
        return la


    def deleteLoginMethods(self):
        self.store.query(LoginMethod, LoginMethod.account == self).deleteFromStore()


    def addLoginMethod(self, localpart, domain, protocol=ANY_PROTOCOL, verified=False, internal=False):
        """
        Add a login method to this account, propogating up or down as necessary
        to site store or user store to maintain consistency.
        """
        # Out takes you west or something
        if self.store.parent is None:
            # West takes you in
            otherStore = self.avatars.open()
            peer = otherStore.findUnique(LoginAccount)
        else:
            # In takes you east
            otherStore = self.store.parent
            subStoreItem = self.store.parent.getItemByID(self.store.idInParent)
            peer = otherStore.findUnique(LoginAccount,
                                         LoginAccount.avatars == subStoreItem)

        # Up and down take you home
        for store, account in [(otherStore, peer), (self.store, self)]:
            store.findOrCreate(LoginMethod,
                               account=account,
                               localpart=localpart,
                               domain=domain,
                               protocol=protocol,
                               verified=verified,
                               internal=internal)


    def setPassword(self, newPassword):
        """
        Set this account's password unconditionally.

        @param newPassword: The new password.

        @return: A deferred firing when the password has been changed.
        """
        self.password = None if newPassword is None else unicode(newPassword)
        return succeed(None)


    def replacePassword(self, currentPassword, newPassword):
        """
        Set this account's password if the current password matches.

        @param currentPassword: The password to match against the current one.
        @param newPassword: The new password.

        @return: A deferred firing when the password has been changed.
        @raise BadCredentials: If the current password did not match.
        """
        if unicode(currentPassword) != self.password:
            return fail(BadCredentials())
        return self.setPassword(newPassword)



def insertUserStore(siteStore, userStorePath):
    """
    Move the SubStore at the indicated location into the given site store's
    directory and then hook it up to the site store's authentication database.

    @type siteStore: C{Store}
    @type userStorePath: C{FilePath}
    """
    # The following may, but does not need to be in a transaction, because it
    # is merely an attempt to guess a reasonable filesystem name to use for
    # this avatar.  The user store being operated on is expected to be used
    # exclusively by this process.
    ls = siteStore.findUnique(LoginSystem)
    unattachedSubStore = Store(userStorePath)
    for lm in unattachedSubStore.query(LoginMethod,
                                       LoginMethod.account == unattachedSubStore.findUnique(LoginAccount),
                                       sort=LoginMethod.internal.descending):
        if ls.accountByAddress(lm.localpart, lm.domain) is None:
            localpart, domain = lm.localpart, lm.domain
            break
    else:
        raise AllNamesConflict()

    unattachedSubStore.close()

    insertLocation = siteStore.newFilePath('account', domain, localpart + '.axiom')
    insertParentLoc = insertLocation.parent()
    if not insertParentLoc.exists():
        insertParentLoc.makedirs()
    if insertLocation.exists():
        raise DatabaseDirectoryConflict()
    userStorePath.moveTo(insertLocation)
    ss = SubStore(store=siteStore, storepath=insertLocation)
    attachedStore = ss.open()
    # migrateUp() manages its own transactions because it interacts with two
    # different stores.
    attachedStore.findUnique(LoginAccount).migrateUp()


def extractUserStore(userAccount, extractionDestination, legacySiteAuthoritative=True):
    """
    Move the SubStore for the given user account out of the given site store
    completely.  Place the user store's database directory into the given
    destination directory.

    @type userAccount: C{LoginAccount}
    @type extractionDestination: C{FilePath}

    @type legacySiteAuthoritative: C{bool}

    @param legacySiteAuthoritative: before moving the user store, clear its
    authentication information, copy that which is associated with it in the
    site store rather than trusting its own.  Currently this flag is necessary
    (and defaults to true) because things like the ClickChronicle
    password-changer gizmo still operate on the site store.

    """
    if legacySiteAuthoritative:
        # migrateDown() manages its own transactions, since it is copying items
        # between two different stores.
        userAccount.migrateDown()
    av = userAccount.avatars
    av.open().close()
    def _():
        # We're separately deleting several Items from the site store, then
        # we're moving some files.  If we cannot move the files, we don't want
        # to delete the items.

        # There is one unaccounted failure mode here: if the destination of the
        # move is on a different mount point, the moveTo operation will fall
        # back to a non-atomic copy; if all of the copying succeeds, but then
        # part of the deletion of the source files fails, we will be left
        # without a complete store in this site store's files directory, but
        # the account Items will remain.  This will cause odd errors on login
        # and at other unpredictable times.  The database is only one file, so
        # we will either remove it all or none of it.  Resolving this requires
        # manual intervention currently: delete the substore's database
        # directory and the account items (LoginAccount and LoginMethods)
        # manually.

        # However, this failure is extremely unlikely, as it would almost
        # certainly indicate a misconfiguration of the permissions on the site
        # store's files area.  As described above, a failure of the call to
        # os.rename(), if the platform's rename is atomic (which it generally
        # is assumed to be) will not move any files and will cause a revert of
        # the transaction which would have deleted the accompanying items.

        av.deleteFromStore()
        userAccount.deleteLoginMethods()
        userAccount.deleteFromStore()
        av.storepath.moveTo(extractionDestination)
    userAccount.store.transact(_)


def upgradeLoginAccount1To2(oldAccount):
    password = oldAccount.password
    if password is not None:
        try:
            password = password.decode('ascii')
        except UnicodeDecodeError:
            password = None

    newAccount = oldAccount.upgradeVersion(
        'login', 1, 2,
        password=password,
        avatars=oldAccount.avatars,
        disabled=oldAccount.disabled)

    def make(s, acc):
        LoginMethod(
            store=s,
            localpart=oldAccount.username,
            domain=oldAccount.domain,
            internal=False,
            protocol=u'email',
            account=acc,
            verified=True)

    make(newAccount.store, newAccount)
    ss = newAccount.avatars.open()
    # create account in substore to represent the user's own record of their
    # password; moves with them during migrations, etc.
    subacc = LoginAccount(store=ss,
                          password=newAccount.password,
                          avatars=ss,
                          disabled=newAccount.disabled)
    make(ss, subacc)

upgrade.registerUpgrader(upgradeLoginAccount1To2, 'login', 1, 2)


class SubStoreLoginMixin:
    def makeAvatars(self, domain, username):
        return SubStore.createNew(self.store, ('account', domain, username + '.axiom'))

class LoginBase:
    """
    I am a database powerup which provides an interface to a collection of
    username/password pairs mapped to user application objects.
    """
    implements(IRealm, ICredentialsChecker)

    credentialInterfaces = (IUsernamePassword, IUsernameHashedPassword)

    powerupInterfaces = (IRealm, ICredentialsChecker)

    def accountByAddress(self, username, domain):
        """
        @type username: C{unicode} without NUL
        @type domain: C{unicode} without NUL
        """
        for account in self.store.query(LoginAccount,
                                     AND(LoginMethod.domain == domain,
                                         LoginMethod.localpart == username,
                                         LoginAccount.disabled == 0,
                                         LoginMethod.account == LoginAccount.storeID)):
            return account


    def addAccount(self, username, domain, password, avatars=None,
                   protocol=u'email', disabled=0, internal=False,
                   verified=True):
        """
        Create a user account, add it to this LoginBase, and return it.

        This method must be called within a transaction in my store.

        @param username: the user's name.

        @param domain: the domain part of the user's name [XXX TODO: this
        really ought to say something about whether it's a Q2Q domain, a SIP
        domain, an HTTP realm, or an email address domain - right now the
        assumption is generally that it's an email address domain, but not
        always]

        @param password: A shared secret.

        @param avatars: (Optional).  A SubStore which, if passed, will be used
        by cred as the target of all adaptations for this user.  By default, I
        will create a SubStore, and plugins can be installed on that substore
        using the powerUp method to provide implementations of cred client
        interfaces.

        @raise DuplicateUniqueItem: if the 'avatars' argument already contains
        a LoginAccount.

        @return: an instance of a LoginAccount, with all attributes filled out
        as they are passed in, stored in my store.
        """

        # unicode(None) == u'None', kids.
        if username is not None:
            username = unicode(username)
        if domain is not None:
            domain = unicode(domain)
        if password is not None:
            password = unicode(password)

        if self.accountByAddress(username, domain) is not None:
            raise DuplicateUser(username, domain)
        if avatars is None:
            avatars = self.makeAvatars(domain, username)

        subStore = avatars.open()

        # create this unconditionally; as the docstring says, we must be run
        # within a transaction, so if something goes wrong in the substore
        # transaction this item's creation will be reverted...
        la = LoginAccount(store=self.store,
                          password=password,
                          avatars=avatars,
                          disabled=disabled)

        def createSubStoreAccountObjects():

            LoginAccount(store=subStore,
                         password=password,
                         disabled=disabled,
                         avatars=subStore)

            la.addLoginMethod(localpart=username,
                              domain=domain,
                              protocol=protocol,
                              internal=internal,
                              verified=verified)

        subStore.transact(createSubStoreAccountObjects)
        return la


    def logoutFactory(self, obj):
        return getattr(obj, 'logout', lambda: None)


    def requestAvatar(self, avatarId, mind, *interfaces):
        if avatarId is ANONYMOUS:
            av = self.store
        else:
            av = self.store.getItemByID(avatarId)
        for interface in interfaces:
            impl = interface(av, None)
            if impl is not None:
                self.loginCount += 1
                log.msg(interface=iaxiom.IStatEvent, name='cred',
                        cred_interface=interface)
                return interface, impl, self.logoutFactory(impl)
        raise NotImplementedError()

    def requestAvatarId(self, credentials):
        try:
            username, domain = credentials.username.split('@', 1)
        except ValueError:
            self.failedLogins += 1
            raise MissingDomainPart(credentials.username)

        username = unicode(username)
        domain = unicode(domain)

        acct = self.accountByAddress(username, domain)
        if acct is not None:
            # Awful hack
            if isinstance(credentials, Preauthenticated):
                return acct.storeID
            elif IUsernameHashedPassword.providedBy(credentials):
                warnings.warn(
                    'Authenticating IUsernameHashedPassword credentials with '
                    'axiom.userbase is deprecated; use IUsernamePassword '
                    'instead', DeprecationWarning)
                if credentials.checkPassword(acct.password):
                    return acct.storeID
                else:
                    self.failedLogins += 1
                    raise BadCredentials()
            else:
                if unicode(credentials.password) == acct.password:
                    return succeed(acct.storeID)
                else:
                    self.failedLogins += 1
                    return fail(BadCredentials())

        self.failedLogins += 1
        raise NoSuchUser(credentials.username)



class LoginSystem(Item, LoginBase, SubStoreLoginMixin):
    schemaVersion = 1
    typeName = 'login_system'

    loginCount = integer(default=0)
    failedLogins = integer(default=0)


def getLoginMethods(store, protocol=None):
    """
    Retrieve L{LoginMethod} items from store C{store}, optionally constraining
    them by protocol
    """
    if protocol is not None:
        comp = OR(LoginMethod.protocol == u'*',
                  LoginMethod.protocol == protocol)
    else:
        comp = None
    return store.query(LoginMethod, comp)


def getAccountNames(store, protocol=None):
    """
    Retrieve account name information about the given database.

    @param store: An Axiom Store representing a user account.  It must
    have been opened through the store which contains its account
    information.

    @return: A generator of two-tuples of (username, domain) which
    refer to the given store.
    """
    return ((meth.localpart, meth.domain) for meth
                in getLoginMethods(store, protocol))


def getDomainNames(store):
    """
    Retrieve a list of all local domain names represented in the given store.
    """
    domains = set()
    domains.update(store.query(
            LoginMethod,
            AND(LoginMethod.internal == True,
                LoginMethod.domain != None)).getColumn("domain").distinct())
    return sorted(domains)




import time
import itertools

from testbase import con, cur
BATCH = 500
for num in itertools.count():
    for x in range(BATCH):
        n = (num * BATCH) + x
        cur.execute("insert into foo values (?, ?)",
                    (n, "string-value-of-"+str(n)))
    con.commit()
    print num, 'write pass complete', time.ctime()



# Copyright (c) 2008 Divmod.  See LICENSE for details.

"""
Tests for the Axiom upgrade system.
"""

import sys, StringIO

from zope.interface import Interface
from zope.interface.verify import verifyObject

from twisted.trial import unittest
from twisted.python import filepath
from twisted.application.service import IService
from twisted.internet.defer import maybeDeferred, succeed
from twisted.python.reflect import namedModule
from twisted.python import log

from axiom.iaxiom import IAxiomaticCommand
from axiom import store, upgrade, item, errors, attributes
from axiom.upgrade import _StoreUpgrade
from axiom.item import declareLegacyItem
from axiom.scripts import axiomatic
from axiom.store import Store
from axiom.substore import SubStore
from axiom.plugins.axiom_plugins import Upgrade
from axiom.test.util import CommandStub



def axiomInvalidate(itemClass):
    """
    Remove the registered item class from the Axiom module system's memory,
    including: the item's current schema, legacy schema declarations, and
    upgraders.

    This makes it possible, for example, to reload a module without Axiom
    complaining about it.

    This API is still in a test module because it is _NOT YET SAFE_ for using
    while databases are open; it does not interact with open databases' caches,
    for example.

    @param itemClass: an Item subclass that you no longer wish to use.
    """
    # Note, be very careful not to use comparison on attributes here.  For
    # example, do not use list.remove(), since it is equality based. -exarkun
    for cascades in attributes._cascadingDeletes.itervalues():
        for i in xrange(len(cascades) - 1, -1, -1):
            if cascades[i].type is itemClass:
                del cascades[i]

    store._typeNameToMostRecentClass.pop(itemClass.typeName, None)

    for (tnam, schever) in item._legacyTypes.keys():
        if tnam == itemClass.typeName:
            item._legacyTypes.pop((tnam, schever))

    for k in upgrade._upgradeRegistry.keys():
        if k[0] == itemClass.typeName:
            upgrade._upgradeRegistry.pop(k)

def axiomInvalidateModule(moduleObject):
    """
    Call L{axiomInvalidate} on all Item subclasses defined in a module.
    """
    for v in moduleObject.__dict__.values():
        if isinstance(v, item.MetaItem):
            axiomInvalidate(v)

schemaModules = []

def loadSchemaModule(name):
    schemaModules.append(namedModule(name))
    result = schemaModules[-1]
    choose(None)
    return result

def choose(module=None):
    """
    Choose among the various "adventurer" modules for upgrade tests.

    @param module: the module object which should next be treated as "current".
    """
    for old in schemaModules:
        axiomInvalidateModule(old)

    if module is not None:
        reload(module)

oldapp = loadSchemaModule('axiom.test.oldapp')

brokenapp = loadSchemaModule('axiom.test.brokenapp')

toonewapp = loadSchemaModule('axiom.test.toonewapp')

morenewapp = loadSchemaModule('axiom.test.morenewapp')

onestepapp = loadSchemaModule('axiom.test.onestepapp')

newapp = loadSchemaModule('axiom.test.newapp')

oldpath = loadSchemaModule('axiom.test.oldpath')

newpath = loadSchemaModule('axiom.test.newpath')

path_postcopy = loadSchemaModule('axiom.test.path_postcopy')

deleteswordapp = loadSchemaModule('axiom.test.deleteswordapp')

class SchemaUpgradeTest(unittest.TestCase):
    def setUp(self):
        self.dbdir = filepath.FilePath(self.mktemp())

    def openStore(self, dbg=False):
        self.currentStore = store.Store(self.dbdir, debug=dbg)
        return self.currentStore


    def closeStore(self):
        """
        Close C{self.currentStore} and discard the reference.  If there is a
        store service running, stop it first.
        """
        service = IService(self.currentStore)
        if service.running:
            result = service.stopService()
        else:
            result = succeed(None)
        def close(ignored):
            self.currentStore.close()
            self.currentStore = None
        result.addCallback(close)
        return result


    def startStoreService(self):
        svc = IService(self.currentStore)
        svc.getServiceNamed("Batch Processing Controller").disownServiceParent()
        svc.startService()



def _logMessagesFrom(f):
    L = []
    log.addObserver(L.append)
    d = maybeDeferred(f)
    def x(ign):
        log.removeObserver(L.append)
        return ign
    return d.addBoth(x).addCallback(lambda ign: L)



def callWithStdoutRedirect(f, *a, **kw):
    """
    Redirect stdout and invoke C{f}.

    @returns: C{(returnValue, stdout})
    """
    output = StringIO.StringIO()
    sys.stdout, stdout = output, sys.stdout
    try:
        result = f(*a, **kw)
    finally:
        sys.stdout = stdout
    return result, output



class SwordUpgradeTest(SchemaUpgradeTest):

    def tearDown(self):
        choose(oldapp)

    def testUnUpgradeableStore(self):
        self._testTwoObjectUpgrade()
        choose(toonewapp)
        self.assertRaises(errors.NoUpgradePathAvailable, self.openStore)

    def testUpgradeWithMissingVersion(self):
        playerID, swordID = self._testTwoObjectUpgrade()
        choose(morenewapp)
        s = self.openStore()
        self.startStoreService()
        def afterUpgrade(result):
            player = s.getItemByID(playerID, autoUpgrade=False)
            sword = s.getItemByID(swordID, autoUpgrade=False)
            self._testPlayerAndSwordState(player, sword)
        return s.whenFullyUpgraded().addCallback(afterUpgrade)


    def test_upgradeSkipVersion(self):
        """
        Verify that an upgrader registered to skip a version can execute properly.
        """
        playerID, swordID = self._testTwoObjectUpgrade()
        choose(onestepapp)
        s = self.openStore()
        self.startStoreService()
        def afterUpgrade(result):
            player = s.getItemByID(playerID, autoUpgrade=False)
            sword = s.getItemByID(swordID, autoUpgrade=False)
            self._testPlayerAndSwordState(player, sword)
        return s.whenFullyUpgraded().addCallback(afterUpgrade)



    def test_loggingAtAppropriateTimes(self):
        """
        Verify that log messages show up when we do upgrade work, but then don't
        when we don't.
        """
        def someLogging(logMessages):
            ok = False
            unrelatedMessages = []
            for msgdict in logMessages:
                msgstr = u''.join(msgdict.get('message', ()))
                if u'finished upgrading' in msgstr:
                    ok = True
                else:
                    unrelatedMessages.append(msgstr)
            self.failUnless(ok, "No messages related to upgrading: %r" % (unrelatedMessages,))
            s = self.openStore()
            def afterUpgrade(noLogMessages):
                for nmsgdict in noLogMessages:
                    mm = u''.join(nmsgdict.get('message', ()))
                    if mm:
                        self.failIfIn(u'finished upgrading', mm)
            self.startStoreService()
            return _logMessagesFrom(s.whenFullyUpgraded
                                    ).addCallback(afterUpgrade)

        return _logMessagesFrom(self.testTwoObjectUpgrade_UseService).addCallback(someLogging)


    def test_basicErrorLogging(self):
        """
        Verify that if an exception is raised in an upgrader, the exception
        will be logged.
        """
        choose(oldapp)
        s = self.openStore()
        swordID = oldapp.Sword(
            store=s,
            name=u'flaming vorpal doom',
            hurtfulness=7).storeID
        self.closeStore()

        choose(brokenapp)
        s = self.openStore()
        self.startStoreService()

        def checkException(ign):
            # It's redundant that the errback is called and the failure is
            # logged.  See #2638.
            loggedErrors = self.flushLoggedErrors(errors.ItemUpgradeError)
            self.assertEqual(len(loggedErrors), 1)
            upgradeError = loggedErrors[0]

            loggedErrors = self.flushLoggedErrors(brokenapp.UpgradersAreBrokenHere)
            self.assertEqual(len(loggedErrors), 1)

            oldType = item.declareLegacyItem(
                oldapp.Sword.typeName,
                oldapp.Sword.schemaVersion, {})
            e = upgradeError.value
            self.assertEqual(e.storeID, swordID)
            self.assertIdentical(e.oldType, oldType)
            self.assertIdentical(e.newType, brokenapp.Sword)

        d = s.whenFullyUpgraded()
        d = self.assertFailure(d, errors.ItemUpgradeError)
        d.addCallback(checkException)
        return d


    def _testTwoObjectUpgrade(self):
        choose(oldapp)
        s = self.openStore()
        self.assertIdentical(
            store._typeNameToMostRecentClass[oldapp.Player.typeName],
            oldapp.Player)

        sword = oldapp.Sword(
            store=s,
            name=u'flaming vorpal doom',
            hurtfulness=7)

        player = oldapp.Player(
            store=s,
            name=u'Milton',
            sword=sword)

        self.closeStore()

        # Perform an adjustment.

        return player.storeID, sword.storeID


    def testTwoObjectUpgrade_OuterFirst(self):
        playerID, swordID = self._testTwoObjectUpgrade()
        player, sword = self._testLoadPlayerFirst(playerID, swordID)
        self._testPlayerAndSwordState(player, sword)


    def testTwoObjectUpgrade_InnerFirst(self):
        playerID, swordID = self._testTwoObjectUpgrade()
        player, sword = self._testLoadSwordFirst(playerID, swordID)
        self._testPlayerAndSwordState(player, sword)


    def testTwoObjectUpgrade_AutoOrder(self):
        playerID, swordID = self._testTwoObjectUpgrade()
        player, sword = self._testAutoUpgrade(playerID, swordID)
        self._testPlayerAndSwordState(player, sword)


    def testTwoObjectUpgrade_UseService(self):
        playerID, swordID = self._testTwoObjectUpgrade()
        choose(newapp)
        s = self.openStore()
        self.startStoreService()

        # XXX *this* test really needs 10 or so objects to play with in order
        # to be really valid...

        def afterUpgrade(result):
            player = s.getItemByID(playerID, autoUpgrade=False)
            sword = s.getItemByID(swordID, autoUpgrade=False)
            self._testPlayerAndSwordState(player, sword)
            # Stop that service we started.
            return IService(s).stopService()

        return s.whenFullyUpgraded().addCallback(afterUpgrade)


    def _testAutoUpgrade(self, playerID, swordID):
        choose(newapp)
        s = self.openStore()

        for dummy in s._upgradeManager.upgradeEverything():
            pass

        player = s.getItemByID(playerID, autoUpgrade=False)
        sword = s.getItemByID(swordID, autoUpgrade=False)

        return player, sword

    def _testLoadPlayerFirst(self, playerID, swordID):
        # Everything old is new again
        choose(newapp)

        s = self.openStore()
        player = s.getItemByID(playerID)
        sword = s.getItemByID(swordID)
        return player, sword

    def _testLoadSwordFirst(self, playerID, swordID):
        choose(newapp)

        s = self.openStore()
        sword = s.getItemByID(swordID)
        player = s.getItemByID(playerID)
        return player, sword

    def _testPlayerAndSwordState(self, player, sword):
        assert not player.__legacy__
        assert not sword.__legacy__
        self.assertEquals(player.name, 'Milton')
        self.failIf(hasattr(player, 'sword'))
        self.assertEquals(sword.name, 'flaming vorpal doom')
        self.assertEquals(sword.damagePerHit, 14)
        self.failIf(hasattr(sword, 'hurtfulness'))
        self.assertEquals(sword.owner.storeID, player.storeID)
        self.assertEquals(type(sword.owner), type(player))
        self.assertEquals(sword.owner, player)
        self.assertEquals(sword.activated, 1)
        self.assertEquals(player.activated, 1)


    def test_multipleLegacyVersions(self):
        """
        If multiple legacy schema versions are present, all of them should be
        upgraded.
        """
        playerID, swordID = self._testTwoObjectUpgrade()

        choose(newapp)
        s = self.openStore()
        self.startStoreService()
        def afterFirstUpgrade(result):
            choose(morenewapp)
            s = self.openStore()
            self.startStoreService()
            return s.whenFullyUpgraded().addCallback(afterSecondUpgrade, s)

        def afterSecondUpgrade(result, store):
            player = store.getItemByID(playerID, autoUpgrade=False)
            sword = store.getItemByID(swordID, autoUpgrade=False)
            self._testPlayerAndSwordState(player, sword)

        d = s.whenFullyUpgraded()
        d.addCallback(lambda ignored: self.closeStore())
        d.addCallback(afterFirstUpgrade)
        return d



class SubStoreCompat(SwordUpgradeTest):
    def setUp(self):
        self.topdbdir = filepath.FilePath(self.mktemp())
        self.subStoreID = None

    def openStore(self):
        self.currentTopStore = store.Store(self.topdbdir)
        if self.subStoreID is not None:
            self.currentSubStore = self.currentTopStore.getItemByID(self.subStoreID).open()
        else:
            ss = SubStore.createNew(self.currentTopStore,
                                    ['sub'])
            self.subStoreID = ss.storeID
            self.currentSubStore = ss.open()
        return self.currentSubStore


    def closeStore(self):
        """
        Close C{self.currentTopStore} and C{self.currentSubStore}.  If there is
        a store service running in C{self.currentTopStore}, stop it first.
        """
        service = IService(self.currentTopStore)
        if service.running:
            result = service.stopService()
        else:
            result = succeed(None)
        def stopped(ignored):
            self.currentSubStore.close()
            self.currentTopStore.close()
            self.currentSubStore = None
            self.currentTopStore = None
        result.addCallback(stopped)
        return result


    def startStoreService(self):
        svc = IService(self.currentTopStore)
        svc.getServiceNamed("Batch Processing Controller").disownServiceParent()
        svc.startService()



class PathUpgrade(SchemaUpgradeTest):
    """
    Tests for items with path attributes, using
    registerAttributeCopyingUpgrader.
    """
    def _runPathUpgrade(self, module):
        """
        Load the 'oldpath' module, then upgrade items created from it to the
        versions in the specified module.
        """
        axiomInvalidateModule(module)
        reload(oldpath)
        self.openStore()
        nfp = self.currentStore.newFilePath("pathname")
        oldpath.Path(store=self.currentStore,
                     thePath=nfp)
        self.closeStore()
        axiomInvalidateModule(oldpath)
        reload(module)
        self.openStore()
        self.startStoreService()
        return nfp, self.currentStore.whenFullyUpgraded()


    def testUpgradePath(self):
        """
        Verify that you can upgrade a path attribute in the simplest possible
        way.
        """
        nfp, d = self._runPathUpgrade(newpath)
        def checkPathEquivalence(n):
            self.assertEquals(
                self.currentStore.findUnique(newpath.Path).thePath.path,
                nfp.path)
        return d.addCallback(checkPathEquivalence)


    def test_postCopy(self):
        """
        Ensure that a post-copy function, if specified to
        registerAttributeCopyingUpgrader, is run after item upgrade.
        """
        nfp, d = self._runPathUpgrade(path_postcopy)
        path2 = nfp.child("foo")
        def checkPath(_):
            self.assertEquals(
                self.currentStore.findUnique(path_postcopy.Path).thePath.path,
                path2.path)
        return d.addCallback(checkPath)


oldcirc = loadSchemaModule('axiom.test.oldcirc')

newcirc = loadSchemaModule('axiom.test.newcirc')

oldobsolete = loadSchemaModule('axiom.test.oldobsolete')

newobsolete = loadSchemaModule('axiom.test.newobsolete')


class IObsolete(Interface):
    """
    Interface representing an undesirable feature.
    """

class DeletionTest(SchemaUpgradeTest):
    def testCircular(self):
        """
        If you access an item, B, through a reference on another item, A, which
        is deleted in the course of B's upgrade, you should still get a
        reference to B.
        """
        reload(oldcirc)
        self.openStore()
        b = oldcirc.B(a=oldcirc.A(store=self.currentStore),
                      store=self.currentStore)
        b.a.b = b

        self.closeStore()

        axiomInvalidateModule(oldcirc)
        reload(newcirc)

        self.openStore()
        origA = self.currentStore.findUnique(newcirc.A)
        origB = origA.b
        secondA = self.currentStore.findUnique(newcirc.A)
        secondB = secondA.b
        self.assertEquals(origB, secondB)
        self.assertNotEqual(origA, secondA)

    def testPowerupsFor(self):
        """
        Powerups deleted during upgrades should be omitted from the results of
        powerupsFor.
        """
        reload(oldobsolete)
        self.openStore()
        o = oldobsolete.Obsolete(store=self.currentStore)
        self.currentStore.powerUp(o, IObsolete)
        # sanity check
        self.assertEquals(IObsolete(self.currentStore), o)
        self.closeStore()

        axiomInvalidateModule(oldobsolete)
        reload(newobsolete)
        self.openStore()
        self.assertEquals(list(self.currentStore.powerupsFor(IObsolete)), [])
        self.closeStore()
        axiomInvalidateModule(newobsolete)

    def testPowerupsAdapt(self):
        """
        Powerups deleted during upgrades should be omitted from the results of
        powerupsFor.
        """
        reload(oldobsolete)
        self.openStore()
        o = oldobsolete.Obsolete(store=self.currentStore)
        self.currentStore.powerUp(o, IObsolete)
        # sanity check
        self.assertEquals(IObsolete(self.currentStore), o)
        self.closeStore()

        axiomInvalidateModule(oldobsolete)
        reload(newobsolete)
        self.openStore()
        self.assertEquals(IObsolete(self.currentStore, None), None)
        self.closeStore()
        axiomInvalidateModule(newobsolete)



two_upgrades_old = loadSchemaModule(
    'axiom.test.upgrade_fixtures.two_upgrades_old')

two_upgrades_new = loadSchemaModule(
    'axiom.test.upgrade_fixtures.two_upgrades_new')

reentrant_old = loadSchemaModule(
    'axiom.test.upgrade_fixtures.reentrant_old')

reentrant_new = loadSchemaModule(
    'axiom.test.upgrade_fixtures.reentrant_new')

override_init_old = loadSchemaModule(
    'axiom.test.upgrade_fixtures.override_init_old')

override_init_new = loadSchemaModule(
    'axiom.test.upgrade_fixtures.override_init_new')

replace_attribute_old = loadSchemaModule(
    'axiom.test.upgrade_fixtures.replace_attribute_old')

replace_attribute_new = loadSchemaModule(
    'axiom.test.upgrade_fixtures.replace_attribute_new')

replace_delete_old = loadSchemaModule(
    'axiom.test.upgrade_fixtures.replace_delete_old')

replace_delete_new = loadSchemaModule(
    'axiom.test.upgrade_fixtures.replace_delete_new')


class DuringUpgradeTests(unittest.TestCase):
    """
    Tests for upgraders' interactions with each other and with the Store while
    an upgrader is running.
    """
    def tearDown(self):
        choose(None)


    dbdir = None
    currentStore = None

    def storeWithVersion(self, chosenModule):
        """
        Open a store with a particular module chosen, closing the old store if
        it was open already.
        """
        choose(chosenModule)
        if self.currentStore is not None:
            self.currentStore.close()
        if self.dbdir is None:
            self.dbdir = filepath.FilePath(self.mktemp())
        self.currentStore = store.Store(self.dbdir)
        return self.currentStore


    def test_upgradeLegacyReference(self):
        """
        Let a and b be two items which are being upgraded, instances of item
        types A and B respectively.  a has a reference attribute, x, which
        points to b.  In A's 1to2 upgrader, newA.x is set to oldA.x, which is
        (at that time) a DummyItem, i.e. an item with __legacy__ set to True.

        This is a regression test for a bug in this scenario where caching was
        too aggressive, and a.x would still refer to a legacy item after the
        upgrade was finished.  After performing this upgrade, a.x should refer
        to a B v2, i.e. an upgraded version of b.
        """
        old = self.storeWithVersion(two_upgrades_old)
        storeID = two_upgrades_old.Referrer(
            store=old,
            referee=two_upgrades_old.Referee(store=old)).storeID

        new = self.storeWithVersion(two_upgrades_new)
        referrer = new.getItemByID(storeID)
        referee = referrer.referee
        self.assertTrue(
            isinstance(referee, two_upgrades_new.Referee),
            "%r is a %r but should be %r" % (
                referee, type(referee), two_upgrades_new.Referee))


    def test_reentrantUpgraderFailure(self):
        """
        If, while an upgrader is running, it triggers its own upgrade, there
        should be a loud failure; it's already hard enough to deal with upgrade
        ordering and querying for legacy items; upgraders cannot reasonably be
        written to be correct in the face of reentrancy.
        """
        old = self.storeWithVersion(reentrant_old)
        storeID = reentrant_old.Simple(store=old).storeID
        new = self.storeWithVersion(reentrant_new)
        self.assertRaises(errors.UpgraderRecursion, new.getItemByID, storeID)
        # A whitebox flourish to make sure our state tracking is correct:
        self.failIf(new._upgradeManager._currentlyUpgrading,
                    "No upgraders should currently be in progress.")


    def test_overridenInitializerInUpgrader(self):
        """
        A subclass of Item which overrides __init__ should be cached by the end
        of Item.__init__, so that logic written by the subclass has normal
        caching semantics.
        """
        old = self.storeWithVersion(override_init_old)
        storeID = override_init_old.Simple(store=old).storeID
        new = self.storeWithVersion(override_init_new)
        upgraded = new.getItemByID(storeID)
        simpleSelf, simpleGotItem = upgraded.verify
        self.assertIdentical(upgraded, simpleSelf)
        self.assertIdentical(upgraded, simpleGotItem)


    def _reentrantReferenceForeignUpgrader(self, oldModule, newModule):
        old = self.storeWithVersion(oldModule)
        storeID = oldModule.Referrer(
            store=old, referee=oldModule.Referee(
                store=old, value=oldModule.OLD_VALUE)).storeID
        new = self.storeWithVersion(newModule)
        referrer = new.getItemByID(storeID)
        upgraded = referrer.referee
        self.assertEqual(
            upgraded.value,
            newModule.NEW_VALUE,
            "Upgraded reference does not have new value.")


    def test_referenceModifiedByForeignUpgrader(self):
        """
        If the value of a reference on an Item requires an upgrade and the
        upgrade replaces the value of the reference with a different Item, then
        evaluating the reference attribute on the referrer should result in the
        new value of the attribute.
        """
        self._reentrantReferenceForeignUpgrader(
            replace_attribute_old, replace_attribute_new)


    def test_cascadingDeletedReferenceModifiedByForeignUpgrader(self):
        """
        If the value of a whenDeleted=CASCADE reference on an Item requires an
        upgrade and the upgrade replaces the value of the reference with a new
        Item and then deletes the old value of the reference, then evaluating
        the reference attribute on the referrer should result in the new value
        of the attribute.
        """
        self._reentrantReferenceForeignUpgrader(
            replace_delete_old, replace_delete_new)



class AxiomaticUpgradeTest(unittest.TestCase):
    """
    L{Upgrade} implements an I{axiomatic} subcommand for synchronously
    upgrading all items in a store.
    """
    def setUp(self):
        """
        Create a temporary on-disk Store and an instance of L{Upgrade}.
        """
        self.dbdir = self.mktemp()
        self.store = store.Store(self.dbdir)


    def tearDown(self):
        """
        Close the temporary Store.
        """
        self.store.close()


    def test_providesCommandInterface(self):
        """
        L{Upgrade} provides L{IAxiomaticCommand}.
        """
        self.assertTrue(verifyObject(IAxiomaticCommand, Upgrade))


    def test_axiomaticSubcommand(self):
        """
        L{Upgrade} is available as a subcommand of I{axiomatic}.
        """
        subCommands = axiomatic.Options().subCommands
        [options] = [cmd[2] for cmd in subCommands if cmd[0] == 'upgrade']
        self.assertIdentical(options, Upgrade)


    def test_successOutput(self):
        """
        Upon successful completion of the upgrade, L{Upgrade} writes a success
        message to stdout.
        """
        cmd = Upgrade()
        cmd.parent = CommandStub(self.store, 'upgrade')
        result, output = callWithStdoutRedirect(cmd.parseOptions, [])
        self.assertEqual(output.getvalue(), 'Upgrade complete\n')


    def test_axiomaticUpgradeEverything(self):
        """
        L{Upgrade.upgradeStore} upgrades all L{Item}s.
        """
        choose(oldapp)
        swordID = oldapp.Sword(
            store=self.store, name=u'broadsword', hurtfulness=5).storeID
        self.store.close()

        choose(deleteswordapp)

        cmd = Upgrade()
        cmd.parent = CommandStub(store.Store(self.dbdir), 'upgrade')

        result, output = callWithStdoutRedirect(
            cmd.parseOptions, ['--count', '100'])

        self.store = store.Store(self.dbdir)
        self.assertRaises(
            KeyError, self.store.getItemByID, swordID, autoUpgrade=False)


    def test_axiomaticUpgradeExceptionBubbling(self):
        """
        Exceptions encountered by L{Upgrade.upgradeStore} are handled and
        re-raised as L{errors.ItemUpgradeError} with attributes indicating
        which L{Item} was being upgraded when the exception occurred.
        """
        choose(oldapp)
        swordID = oldapp.Sword(
            store=self.store, name=u'longsword', hurtfulness=4).storeID
        self.store.close()

        choose(brokenapp)
        self.store = store.Store(self.dbdir)

        cmd = Upgrade()
        cmd.parent = CommandStub(self.store, 'upgrade')
        cmd.count = 100

        err = self.assertRaises(
            errors.ItemUpgradeError,
            callWithStdoutRedirect, cmd.upgradeStore, self.store)

        self.assertTrue(
            err.originalFailure.check(brokenapp.UpgradersAreBrokenHere))

        oldType = item.declareLegacyItem(
            oldapp.Sword.typeName,
            oldapp.Sword.schemaVersion, {})
        self.assertEqual(err.storeID, swordID)
        self.assertIdentical(err.oldType, oldType)
        self.assertIdentical(err.newType, brokenapp.Sword)


    def test_axiomaticUpgradePerformFails(self):
        """
        If an exception occurs while upgrading items, L{Upgrade.postOptions}
        reports the item and schema version for which it occurred and returns
        without exception.
        """
        choose(oldapp)
        swordID = oldapp.Sword(
            store=self.store, name=u'rapier', hurtfulness=3).storeID
        self.store.close()

        choose(brokenapp)
        self.store = store.Store(self.dbdir)

        cmd = Upgrade()
        cmd.parent = CommandStub(self.store, 'upgrade')

        result, output = callWithStdoutRedirect(
            cmd.parseOptions, ['--count', '100'])
        lines = output.getvalue().splitlines()

        # Ensure that the original error is output.
        self.assertEqual(lines[0], 'Upgrader error:')
        self.assertTrue(len(lines) > 2)

        oldType = oldapp.Sword
        newType = store._typeNameToMostRecentClass[oldType.typeName]
        msg = cmd.errorMessageFormat % (
            oldType.typeName, swordID, oldType.schemaVersion,
            newType.schemaVersion)
        self.assertTrue(lines[-1].startswith(msg))


    def test_upgradeStoreRecursing(self):
        """
        L{Upgrade} upgrades L{Item}s in substores.
        """
        choose(oldapp)

        ss1 = SubStore.createNew(self.store, ['a'])
        ss2 = SubStore.createNew(self.store, ['b'])

        swordIDs = [
            (ss1.storeID, oldapp.Sword(store=ss1.open(), name=u'foo').storeID),
            (ss2.storeID, oldapp.Sword(store=ss2.open(), name=u'bar').storeID)]

        del ss1, ss2
        self.store.close()

        choose(deleteswordapp)
        self.store = store.Store(self.dbdir)

        cmd = Upgrade()
        cmd.parent = CommandStub(self.store, 'upgrade')

        callWithStdoutRedirect(cmd.parseOptions, [])

        for (ssid, swordID) in swordIDs:
            self.assertRaises(
                KeyError,
                self.store.getItemByID(ssid).open().getItemByID, swordID)


class StoreUpgradeTests(unittest.TestCase):
    """
    Tests for L{upgrade._StoreUpgrade}.
    """
    def setUp(self):
        self.store = Store()
        self._upgrader = _StoreUpgrade(self.store)

    def test_queueMultipleVersions(self):
        """
        If multiple schema versions are queued for upgrade, upgrades should be
        attempted for all of them (but only attempted once per version).
        """
        legacy1 = declareLegacyItem('test_type', 1, {})
        legacy2 = declareLegacyItem('test_type', 2, {})
        self._upgrader.queueTypeUpgrade(legacy1)
        self._upgrader.queueTypeUpgrade(legacy2)
        self._upgrader.queueTypeUpgrade(legacy2)
        self.assertEqual(len(self._upgrader._oldTypesRemaining), 2)




"""
Benchmark creation of a large number of simple Items in a transaction.
"""

from epsilon.scripts import benchmark

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, text

class AB(Item):
    a = integer()
    b = text()

def main():
    s = Store("TEMPORARY.axiom")
    def txn():
        for x in range(10000):
            AB(a=x, b=unicode(x), store=s)

    benchmark.start()
    s.transact(txn)
    benchmark.stop()


if __name__ == '__main__':
    main()



from twisted.trial.unittest import TestCase

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, AND, OR

class ThingsWithIntegers(Item):
    schemaVersion = 1
    typeName = 'axiom_test_thing_with_integers'

    a = integer()
    b = integer()


class NotARealThing(Item):
    schemaVersion = 1
    typeName = 'axiom_test_never_created_item'

    irrelevantAttribute = integer()

    def __init__(self, **kw):
        raise NotImplementedError("You cannot create things that are not real!")


class TestCountQuery(TestCase):

    RANGE = 10
    MIDDLE = 5


    def assertCountEqualsQuery(self, item, cond = None):
        self.assertEquals(self.store.count(item, cond),
                          len(list(self.store.query(item, cond))),
                          'count and len(list(query)) not equals: %r,%r'%(item, cond))

    def setUp(self):
        self.store = Store()
        def populate():
            for i in xrange(self.RANGE):
                for j in xrange(self.RANGE):
                    ThingsWithIntegers(store = self.store, a = i, b = j)
        self.store.transact(populate)

    def testBasicCount(self):
        self.assertCountEqualsQuery(ThingsWithIntegers)

    def testSimpleConditionCount(self):
        self.assertCountEqualsQuery(ThingsWithIntegers,
                                    ThingsWithIntegers.a > self.MIDDLE)

    def testTwoFieldsConditionCount(self):
        self.assertCountEqualsQuery(ThingsWithIntegers,
                                    ThingsWithIntegers.a == ThingsWithIntegers.b)

    def testANDConditionCount(self):
        self.assertCountEqualsQuery(ThingsWithIntegers,
                                    AND(ThingsWithIntegers.a > self.MIDDLE, ThingsWithIntegers.b < self.MIDDLE))

    def testORConditionCount(self):
        self.assertCountEqualsQuery(ThingsWithIntegers,
                                    OR(ThingsWithIntegers.a > self.MIDDLE, ThingsWithIntegers.b < self.MIDDLE))

    def testEmptyResult(self):
        self.assertCountEqualsQuery(ThingsWithIntegers,
                                    ThingsWithIntegers.a == self.RANGE + 3)

    def testNonExistentTable(self):
        self.assertCountEqualsQuery(NotARealThing,
                                    NotARealThing.irrelevantAttribute == self.RANGE + 3)



# -*- test-case-name: axiom.test.test_upgrading.DeletionTest.testPowerups -*-

from axiom.item import Item
from axiom.attributes import integer

class Obsolete(Item):
    """
    This is a stub placeholder so that axiomInvalidateModule will invalidate
    the appropriate typeName; it's probably bad practice to declare recent
    versions of deleted portions of the schema, but that's not what this is
    testing.
    """
    typeName = 'test_upgrading_obsolete'
    nothing = integer()
    schemaVersion = 2

from axiom.upgrade import registerUpgrader

def obsolete1toNone(oldObsolete):
    oldObsolete.deleteFromStore()
    return None

registerUpgrader(obsolete1toNone, 'test_upgrading_obsolete', 1, 2)



# -*- test-case-name: axiomatic.test.test_axiomatic -*-
from zope.interface import alsoProvides, noLongerProvides

import os
import sys
import glob
import errno
import signal

from twisted.plugin import IPlugin, getPlugins
from twisted.python import usage
from twisted.python.runtime import platform
from twisted.scripts import twistd

from axiom.iaxiom import IAxiomaticCommand

class AxiomaticSubCommandMixin(object):
    store = property(lambda self: self.parent.getStore())

    def decodeCommandLine(self, cmdline):
        """Turn a byte string from the command line into a unicode string.
        """
        codec = getattr(sys.stdin, 'encoding', None) or sys.getdefaultencoding()
        return unicode(cmdline, codec)



class _AxiomaticCommandMeta(type):
    """
    Metaclass for L{AxiomaticCommand}.

    This serves to make subclasses provide L{IPlugin} and L{IAxiomaticCommand}.
    """
    def __new__(cls, name, bases, attrs):
        newcls = type.__new__(cls, name, bases, attrs)
        alsoProvides(newcls, IPlugin, IAxiomaticCommand)
        return newcls



class AxiomaticSubCommand(usage.Options, AxiomaticSubCommandMixin):
    """
    L{twisted.python.usage.Options} subclass for Axiomatic sub commands.
    """



class AxiomaticCommand(usage.Options, AxiomaticSubCommandMixin):
    """
    L{twisted.python.usage.Options} subclass for Axiomatic plugin commands.

    Subclass this to have your class automatically provide the necessary
    interfaces to be picked up by axiomatic.
    """
    __metaclass__ = _AxiomaticCommandMeta

noLongerProvides(AxiomaticCommand, IPlugin)
noLongerProvides(AxiomaticCommand, IAxiomaticCommand)



class PIDMixin:

    def _sendSignal(self, signal):
        if platform.isWindows():
            raise usage.UsageError("You can't send signals on Windows (XXX TODO)")
        dbdir = self.parent.getStoreDirectory()
        serverpid = int(file(os.path.join(dbdir, 'run', 'axiomatic.pid')).read())
        os.kill(serverpid, signal)
        return serverpid

    def signalServer(self, signal):
        try:
            return self._sendSignal(signal)
        except (OSError, IOError) as e:
            if e.errno in (errno.ENOENT, errno.ESRCH):
                raise usage.UsageError('There is no server running from the Axiom database %r.' % (self.parent.getStoreDirectory(),))
            else:
                raise


class Stop(usage.Options, PIDMixin):
    def postOptions(self):
        self.signalServer(signal.SIGINT)



class Status(usage.Options, PIDMixin):
    def postOptions(self):
        dbdir = self.parent.getStoreDirectory()
        serverpid = self.signalServer(0)
        print 'A server is running from the Axiom database %r, PID %d.' % (dbdir, serverpid)



class Start(twistd.ServerOptions):
    run = staticmethod(twistd.run)

    def subCommands():
        raise AttributeError()
    subCommands = property(subCommands)


    def getArguments(self, store, args):
        run = store.dbdir.child("run")
        logs = run.child("logs")
        handleLogfile = True
        handlePidfile = True

        for arg in args:
            if arg.startswith("--logfile=") or arg in (
                "-l", "--logfile", "-n", "--nodaemon"
            ):
                handleLogfile = False
            elif arg.startswith("--pidfile=") or arg == "--pidfile":
                handlePidfile = False

        if handleLogfile:
            if not logs.exists():
                logs.makedirs()
            args.extend(["--logfile", logs.child("axiomatic.log").path])

        if not platform.isWindows() and handlePidfile:
            args.extend(["--pidfile", run.child("axiomatic.pid").path])
        args.extend(["axiomatic-start", "--dbdir", store.dbdir.path])
        if store.journalMode is not None:
            args.extend(['--journal-mode', store.journalMode.encode('ascii')])
        return args


    def parseOptions(self, args):
        if "--help" in args:
            self.opt_help()
        else:
            # If a reactor is being selected, it must be done before the store
            # is opened, since that may execute arbitrary application code
            # which may in turn install the default reactor.
            for index, arg in enumerate(args):
                if arg in ("--reactor", "-r"):
                    shortName = args[index + 1]
                    del args[index:index + 2]
                    self.opt_reactor(shortName)
                    break
                elif arg.startswith("--reactor="):
                    shortName = arg.split("=")[1]
                    del args[index]
                    self.opt_reactor(shortName)
                    break
            sys.argv[1:] = self.getArguments(self.parent.getStore(), args)
            self.run()



class Options(usage.Options):
    def subCommands():
        def get(self):
            yield ('start', None, Start, 'Launch the given Axiom database')
            if not platform.isWindows():
                yield ('stop', None, Stop, 'Stop the server running from the given Axiom database')
                yield ('status', None, Status, 'Report whether a server is running from the given Axiom database')

            from axiom import plugins
            for plg in getPlugins(IAxiomaticCommand, plugins):
                try:
                    yield (plg.name, None, plg, plg.description)
                except AttributeError:
                    raise RuntimeError("Maldefined plugin: %r" % (plg,))
        return get,
    subCommands = property(*subCommands())

    optParameters = [
        ('dbdir', 'd', None, 'Path containing axiom database to configure/create'),
        ('journal-mode', None, None, 'SQLite journal mode to set'),
        ]

    optFlags = [
        ('debug', 'b', 'Enable Axiom-level debug logging')]

    store = None

    def usedb(self, potentialdb):
        yn = raw_input("Use database %r? (Y/n) " % (potentialdb,))
        if yn.lower() in ('y', 'yes', ''):
            self['dbdir'] = potentialdb
        else:
            raise usage.UsageError('Select another database with the -d option, then.')

    def getStoreDirectory(self):
        if self['dbdir'] is None:
            possibilities = glob.glob('*.axiom')
            if len(possibilities) > 1:
                raise usage.UsageError(
                    "Multiple databases found here, please select one with "
                    "the -d option: %s" % (' '.join(possibilities),))
            elif len(possibilities) == 1:
                self.usedb(possibilities[0])
            else:
                self.usedb(self.subCommand + '.axiom')
        return self['dbdir']

    def getStore(self):
        from axiom.store import Store
        jm = self['journal-mode']
        if jm is not None:
            jm = jm.decode('ascii')
        if self.store is None:
            self.store = Store(
                self.getStoreDirectory(), debug=self['debug'], journalMode=jm)
        return self.store


    def postOptions(self):
        if self.store is not None:
            self.store.close()


def main(argv=None):
    o = Options()
    try:
        o.parseOptions(argv)
    except usage.UsageError as e:
        raise SystemExit(str(e))



# -*- test-case-name: axiom.test -*-

from twisted.cred.error import UnauthorizedLogin


class TimeoutError(Exception):
    """
    A low-level SQL operation timed out.

    @ivar statement: The SQL statement which timed out.
    @ivar timeout: The timeout, in seconds, which was exceeded.
    @ivar underlying: The backend exception which signaled this, or None.
    """
    def __init__(self, statement, timeout, underlying):
        Exception.__init__(self, statement, timeout, underlying)
        self.statement = statement
        self.timeout = timeout
        self.underlying = underlying



class BadCredentials(UnauthorizedLogin):
    pass



class NoSuchUser(UnauthorizedLogin):
    pass



class MissingDomainPart(NoSuchUser):
    """
    Raised when a login is attempted with a username which consists of only
    a local part.  For example, "testuser" instead of "testuser@example.com".
    """


class DuplicateUser(Exception):
    pass



class CannotOpenStore(RuntimeError):
    """
    There is a problem such that the store cannot be opened.
    """



class NoUpgradePathAvailable(CannotOpenStore):
    """
    No upgrade path is available, so the store cannot be opened.
    """



class NoCrossStoreReferences(AttributeError):
    """
    References are not allowed between items within different Stores.
    """



class SQLError(RuntimeError):
    """
    Axiom internally generated some bad SQL.
    """
    def __init__(self, sql, args, underlying):
        RuntimeError.__init__(self, sql, args, underlying)
        self.sql, self.args, self.underlying = self.args

    def __str__(self):
        return "<SQLError: %r(%r) caused %s: %s>" % (
            self.sql, self.args,
            self.underlying.__class__, self.underlying)


class TableAlreadyExists(SQLError):
    """
    Axiom internally created a table at the same time as another database.
    (User code should not need to catch this exception.)
    """



class UnknownItemType(Exception):
    """
    Can't load an item: it's of a type that I don't see anywhere in Python.
    """



class SQLWarning(Warning):
    """
    Axiom internally generated some CREATE TABLE SQL that ... probably wasn't bad
    """



class TableCreationConcurrencyError(RuntimeError):
    """
    Woah, this is really bad.  If you can get this please tell us how.
    """



class DuplicateUniqueItem(KeyError):
    """
    Found 2 or more of an item which is supposed to be unique.
    """



class ItemNotFound(KeyError):
    """
    Did not find even 1 of an item which was supposed to exist.
    """



class ItemClassesOnly(TypeError):
    """
    An object was passed to a method that wasn't a subclass of Item.
    """


class ChangeRejected(Exception):
    """
    Raised when an attempt is made to change the database at a time when
    database changes are disallowed for reasons of consistency.

    This is raised when an application-level callback (for example, committed)
    attempts to change database state.
    """

class DependencyError(Exception):
    """
    Raised when an item can't be installed or uninstalled.
    """

class DeletionDisallowed(ValueError):
    """
    Raised when an attempt is made to delete an item that is referred to by
    reference attributes with whenDeleted == DISALLOW.
    """

class DataIntegrityError(RuntimeError):
    """
    Data integrity seems to have been lost.
    """



class BrokenReference(DataIntegrityError):
    """
    A reference to a nonexistent item was detected when this should be
    impossible.
    """



class UpgraderRecursion(RuntimeError):
    """
    Upgraders are not allowed to recurse.
    """



class ItemUpgradeError(RuntimeError):
    """
    Attempting to upgrade an Item resulted in an error.

    @ivar originalFailure: The failure that caused the item upgrade to fail
    @ivar storeID: Store ID of the item that failed to upgrade
    @ivar oldType: The type of the item being upgraded
    @ivar newType: The type the item should've been upgraded to
    """
    def __init__(self, originalFailure, storeID, oldType, newType):
        RuntimeError.__init__(self, originalFailure, storeID, oldType, newType)
        self.originalFailure = originalFailure
        self.storeID = storeID
        self.oldType = oldType
        self.newType = newType



class UnsatisfiedRequirement(AttributeError):
    """
    A requirement described by a L{axiom.dependency.requiresFromSite} was not
    satisfied by the database, and could not be satisfied automatically at
    runtime by a default factory.
    """




"""
Tests for the upgrade of Scheduler from version 1 to version 2, in which it was
largely supplanted by L{_SiteScheduler}.
"""

from axiom.iaxiom import IScheduler
from axiom.scheduler import Scheduler, _SiteScheduler
from axiom.test.historic.stubloader import StubbedTest


class SchedulerUpgradeTests(StubbedTest):
    def test_powerdown(self):
        """
        The L{Scheduler} created by the stub is powered down by the upgrade and
        adapting the L{Store} to L{IScheduler} succeeds with an instance of
        L{_SiteScheduler}.
        """
        scheduler = self.store.findUnique(Scheduler)
        self.assertEquals(list(self.store.interfacesFor(scheduler)), [])
        self.assertIsInstance(IScheduler(self.store), _SiteScheduler)



# Copyright 2008 Divmod, Inc.  See LICENSE file for details

"""
Axiom plugins for Twisted.
"""

from zope.interface import classProvides

from twisted.plugin import IPlugin, getPlugins
from twisted.python.usage import Options
from twisted.application.service import IServiceMaker, IService, Service


class _CheckSystemVersion(Service):
    """
    A service which, when started, updates the stored version information in a
    store.

    @ivar store: The L{Store} in which to update version information.
    """
    def __init__(self, store):
        self.store = store


    def startService(self):
        from axiom.listversions import checkSystemVersion
        checkSystemVersion(self.store)



class AxiomaticStart(object):
    """
    L{IServiceMaker} plugin which gets an L{IService} from an Axiom store.
    """
    classProvides(IPlugin, IServiceMaker)

    tapname = "axiomatic-start"
    description = "Run an Axiom database (use 'axiomatic start' instead)"

    class options(Options):
        optParameters = [
            ('dbdir', 'd', None, 'Path containing Axiom database to start'),
            ('journal-mode', None, None, 'SQLite journal mode to set'),
            ]

        optFlags = [('debug', 'b', 'Enable Axiom-level debug logging')]


    def makeService(cls, options):
        """
        Create an L{IService} for the database specified by the given
        configuration.
        """
        from axiom.store import Store
        jm = options['journal-mode']
        if jm is not None:
            jm = jm.decode('ascii')
        store = Store(options['dbdir'], debug=options['debug'], journalMode=jm)
        service = IService(store)
        _CheckSystemVersion(store).setServiceParent(service)
        return service
    makeService = classmethod(makeService)


__all__ = ['AxiomaticStart']




from twisted.trial import unittest
from axiom.slotmachine import SetOnce, Attribute, SlotMachine, SchemaMachine

class A(SlotMachine):
    slots = ['a', 'initialized']

class B(SchemaMachine):
    test = Attribute()
    initialized = SetOnce()
    other = SetOnce(default=None)
    nondescriptor = 'readable'
    method = lambda self: self

class Bsub(B):
    pass

class C(object):
    __slots__ = ['a', 'b',
                 'c', 'initialized']

class D:
    def activate(self):
        self.initialized = 1
        self.test = 2
        self.a = 3
        self.b = 4
        self.c = 5

class E(object):
    pass

class X(B, A, C, D, E):
    pass

class Y(Bsub):
    blah = SetOnce()

class ClassWithDefault:
    x = 1

class DefaultTest(SchemaMachine, ClassWithDefault):
    x = Attribute()

class Decoy(ClassWithDefault):
    pass

class DecoyDefault(Decoy, DefaultTest):
    pass

class DefaultOverride(DefaultTest):
    x = 5

class SlotMachineTest(unittest.TestCase):

    def assertBSchema(self, b):
        """
        Test that the given instance conforms to L{B}'s schema.
        """
        b.test = 1
        b.test = 2

        self.assertEquals(b.test, 2)
        self.assertRaises(AttributeError, setattr, b, 'nottest', 'anything')
        self.assertRaises(AttributeError, getattr, b, 'nottest')
        self.assertEquals(b.other, None)
        b.other = 7
        self.assertEquals(b.other, 7)
        self.assertRaises(AttributeError, setattr, b, 'other', 'anything')

        self.assertEquals(b.nondescriptor, 'readable')
        err = self.assertRaises(AttributeError,
                                setattr, b, 'nondescriptor', 'writable')
        self.assertEquals(str(err),
                          "%r can't set attribute 'nondescriptor'"
                          % (type(b).__name__,))
        self.assertEquals(b.nondescriptor, 'readable')

        self.assertEquals(b.method(), b)
        err = self.assertRaises(AttributeError,
                                setattr, b, 'method', lambda: 5)
        self.assertEquals(str(err),
                          "%r can't set attribute 'method'"
                          % (type(b).__name__,))
        self.assertEquals(b.method(), b)

    def testAttributesNotAllowed(self):
        b = B()
        self.assertBSchema(b)

    def testTrivialSubclass(self):
        b = Bsub()
        self.assertBSchema(b)

    def testSetOnce(self):
        b = B()
        b.initialized = 1
        self.assertRaises(AttributeError, setattr, b, 'initialized', 2)
        self.assertEquals(b.initialized, 1)


    def testClassicMixin(self):
        x = X()
        x.activate()

        self.assertRaises(AttributeError, setattr, x, 'initialized', 2)
        self.assertRaises(AttributeError, setattr, x, 'nottest', 'anything')
        self.assertRaises(AttributeError, getattr, x, 'nottest')


    def testAttributesTraverseDeepHierarchy(self):
        y = Y()
        self.assertBSchema(y)


    def test_baseDefault(self):
        """
        L{DefaultTest.x} should take precedence over L{ClassWithDefault.x}.
        """
        dt = DefaultTest()
        # self.failUnless('x' in dt.__slots__, 'x not in '+repr(dt.__slots__) )
        dt.x = 2


    def test_decoyDefault(self):
        """
        Same as L{test_baseDefault}, but with a decoy subclass.
        """
        d = DecoyDefault()
        d.x = 2


    def test_descriptorOverride(self):
        """
        L{DefaultOverride.x} should take precedence over L{DefaultTest.x}
        and prevent the I{x} attribute from being set.
        """
        d = DefaultOverride()
        err = self.assertRaises(AttributeError, setattr, d, 'x', 23)
        self.assertEquals(str(err), "'DefaultOverride' can't set attribute 'x'")
        self.assertEquals(d.x, 5)



# -*- test-case-name: axiom.test.historic.test_textlist -*-


from axiom.item import Item
from axiom.attributes import textlist
from axiom.test.historic.stubloader import saveStub

class Dummy(Item):
    typeName = 'axiom_textlist_dummy'
    schemaVersion = 1

    attribute = textlist(doc="a textlist")



def createDatabase(s):
    """
    Populate the given Store with some Dummy items.
    """
    Dummy(store=s, attribute=[u'foo', u'bar'])



if __name__ == '__main__':
    saveStub(createDatabase, 11858)



import sys

from axiom import store
from twisted.python import filepath

def main(storePath, itemID):

    assert 'axiom.test.itemtest' not in sys.modules, "Test is invalid."

    st = store.Store(filepath.FilePath(storePath))
    item = st.getItemByID(itemID)
    print item.plain

if __name__ == '__main__':
    main(storePath=sys.argv[1], itemID=int(sys.argv[2]))




from twisted.trial import unittest

from axiom.store import Store
from axiom.tags import Catalog

from axiom.item import Item
from axiom.attributes import text

class Gizmo(Item):
    typeName = 'test_gizmo'
    schemaVersion = 1
    name = text()



class TagTestCase(unittest.TestCase):

    def testTagging(self):
        s = Store()
        c = Catalog(store=s)
        g1 = Gizmo(store=s, name=u'one')
        g2 = Gizmo(store=s, name=u'two')

        c.tag(g1, u'single')
        c.tag(g1, u'multi')
        c.tag(g2, u'multi')
        c.tag(g1, u'multi')

        self.assertEquals(list(c.tagsOf(g1)),
                          [u'single', u'multi'])
        self.assertEquals(list(c.tagsOf(g2)),
                          [u'multi'])

        self.assertEquals(list(c.objectsIn(u'single')), [g1])
        self.assertEquals(list(c.objectsIn(u'multi')), [g1, g2])




"""
Benchmark batch creation of a large number of simple Items in a transaction.
"""

from epsilon.scripts import benchmark

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, text

class AB(Item):
    a = integer()
    b = text()

def main():
    s = Store("TEMPORARY.axiom")
    benchmark.start()
    rows = [(x, unicode(x)) for x in xrange(10000)]
    s.transact(lambda: s.batchInsert(AB, (AB.a, AB.b), rows))
    benchmark.stop()


if __name__ == '__main__':
    main()




from axiom.test.historic.stubloader import StubbedTest
from axiom.test.historic.stub_textlist import Dummy

class TextlistTransitionTest(StubbedTest):
    def test_transition(self):
        """
        Test that the textlist survives the encoding transition intact.
        """
        d = self.store.findUnique(Dummy)
        self.assertEquals(d.attribute, [u'foo', u'bar'])



# test-case-name: axiom.test.historic.test_subscheduler1to2

"""
Database creator for the test for the upgrade of SubScheduler from version 1 to
version 2.
"""

from axiom.test.historic.stubloader import saveStub

from axiom.substore import SubStore
from axiom.scheduler import SubScheduler
from axiom.dependency import installOn


def createDatabase(store):
    sub = SubStore.createNew(store, ["substore"]).open()
    installOn(SubScheduler(store=sub), sub)


if __name__ == '__main__':
    saveStub(createDatabase, 17606)



# -*- test-case-name: axiom.test.test_upgrading.DuringUpgradeTests.test_overridenInitializerInUpgrader -*-

from axiom.attributes import integer, inmemory
from axiom.item import Item, normalize
from axiom.upgrade import registerAttributeCopyingUpgrader

class Simple(Item):
    # Don't import the old version, otherwise its schema will get loaded.  This
    # is valid in the upgrade tests, but not at other times. -exarkun
    typeName = normalize(
        "axiom.test.upgrade_fixtures.override_init_old.Simple")
    schemaVersion = 2

    dummy = integer()
    verify = inmemory()

    def __init__(self, **stuff):
        """
        Override Item's __init__ to re-retrieve this Item from the store.
        """
        Item.__init__(self, **stuff)
        self.verify = (self, self.store.getItemByID(self.storeID))

registerAttributeCopyingUpgrader(Simple, 1, 2)



# -*- test-case-name: axiom.test.test_scheduler -*-

"""
Timed event scheduling for Axiom databases.

With this module, applications can schedule an L{Item} to have its C{run} method
called at a particular point in the future.  This call will happen even if the
process which initially schedules it exits and the database is later re-opened
by another process (of course, if the scheduled time comes and goes while no
process is using the database, then the call will be delayed until some process
opens the database and starts its services).

This module contains two implementations of the L{axiom.iaxiom.IScheduler}
interface, one for site stores and one for sub-stores.  Items can only be
scheduled using an L{IScheduler} implementations from the store containing the
item.  This means a typical way to schedule an item to be run is::

    IScheduler(item.store).schedule(item, when)

The scheduler service can also be retrieved from the site store's service
collection by name::

    IServiceCollection(siteStore).getServiceNamed(SITE_SCHEDULER)
"""

import warnings

from zope.interface import implements

from twisted.internet import reactor

from twisted.application.service import IService, Service
from twisted.python import log, failure

from epsilon.extime import Time

from axiom.iaxiom import IScheduler
from axiom.item import Item, declareLegacyItem
from axiom.attributes import AND, timestamp, reference, integer, inmemory, bytes
from axiom.dependency import uninstallFrom
from axiom.upgrade import registerUpgrader
from axiom.substore import SubStore

VERBOSE = False

SITE_SCHEDULER = u"Site Scheduler"


class TimedEventFailureLog(Item):
    typeName = 'timed_event_failure_log'
    schemaVersion = 1

    desiredTime = timestamp()
    actualTime = timestamp()

    runnable = reference()
    traceback = bytes()


class TimedEvent(Item):
    typeName = 'timed_event'
    schemaVersion = 1

    time = timestamp(indexed=True)
    runnable = reference()

    running = inmemory(doc='True if this event is currently running.')

    def activate(self):
        self.running = False


    def _rescheduleFromRun(self, newTime):
        """
        Schedule this event to be run at the indicated time, or if the
        indicated time is None, delete this event.
        """
        if newTime is None:
            self.deleteFromStore()
        else:
            self.time = newTime


    def invokeRunnable(self):
        """
        Run my runnable, and reschedule or delete myself based on its result.
        Must be run in a transaction.
        """
        runnable = self.runnable
        if runnable is None:
            self.deleteFromStore()
        else:
            try:
                self.running = True
                newTime = runnable.run()
            finally:
                self.running = False
            self._rescheduleFromRun(newTime)


    def handleError(self, now, failureObj):
        """ An error occurred running my runnable.  Check my runnable for an
        error-handling method called 'timedEventErrorHandler' that will take
        the given failure as an argument, and execute that if available:
        otherwise, create a TimedEventFailureLog with information about what
        happened to this event.

        Must be run in a transaction.
        """
        errorHandler = getattr(self.runnable, 'timedEventErrorHandler', None)
        if errorHandler is not None:
            self._rescheduleFromRun(errorHandler(self, failureObj))
        else:
            self._defaultErrorHandler(now, failureObj)


    def _defaultErrorHandler(self, now, failureObj):
        TimedEventFailureLog(store=self.store,
                             desiredTime=self.time,
                             actualTime=now,
                             runnable=self.runnable,
                             traceback=failureObj.getTraceback())
        self.deleteFromStore()



class _WackyControlFlow(Exception):
    def __init__(self, eventObject, failureObject):
        Exception.__init__(self, "User code failed during timed event")
        self.eventObject = eventObject
        self.failureObject = failureObject


MAX_WORK_PER_TICK = 10

class SchedulerMixin:
    def _oneTick(self, now):
        theEvent = self._getNextEvent(now)
        if theEvent is None:
            return False
        try:
            theEvent.invokeRunnable()
        except:
            raise _WackyControlFlow(theEvent, failure.Failure())
        self.lastEventAt = now
        return True


    def _getNextEvent(self, now):
        # o/` gonna party like it's 1984 o/`
        theEventL = list(self.store.query(TimedEvent,
                                          TimedEvent.time <= now,
                                          sort=TimedEvent.time.ascending,
                                          limit=1))
        if theEventL:
            return theEventL[0]


    def tick(self):
        now = self.now()
        self.nextEventAt = None
        workBeingDone = True
        workUnitsPerformed = 0
        errors = 0
        while workBeingDone and workUnitsPerformed < MAX_WORK_PER_TICK:
            try:
                workBeingDone = self.store.transact(self._oneTick, now)
            except _WackyControlFlow as wcf:
                self.store.transact(wcf.eventObject.handleError, now, wcf.failureObject)
                log.err(wcf.failureObject)
                errors += 1
                workBeingDone = True
            if workBeingDone:
                workUnitsPerformed += 1
        x = list(self.store.query(TimedEvent, sort=TimedEvent.time.ascending, limit=1))
        if x:
            self._transientSchedule(x[0].time, now)
        if errors or VERBOSE:
            log.msg("The scheduler ran %(eventCount)s events%(errors)s." % dict(
                    eventCount=workUnitsPerformed,
                    errors=(errors and (" (with %d errors)" % (errors,))) or ''))


    def schedule(self, runnable, when):
        TimedEvent(store=self.store, time=when, runnable=runnable)
        self._transientSchedule(when, self.now())


    def reschedule(self, runnable, fromWhen, toWhen):
        for evt in self.store.query(TimedEvent,
                                    AND(TimedEvent.time == fromWhen,
                                        TimedEvent.runnable == runnable)):
            evt.time = toWhen
            self._transientSchedule(toWhen, self.now())
            break
        else:
            raise ValueError("%r is not scheduled to run at %r" % (runnable, fromWhen))


    def unscheduleFirst(self, runnable):
        """
        Remove from given item from the schedule.

        If runnable is scheduled to run multiple times, only the temporally first
        is removed.
        """
        for evt in self.store.query(TimedEvent, TimedEvent.runnable == runnable, sort=TimedEvent.time.ascending):
            evt.deleteFromStore()
            break


    def unscheduleAll(self, runnable):
        for evt in self.store.query(TimedEvent, TimedEvent.runnable == runnable):
            evt.deleteFromStore()


    def scheduledTimes(self, runnable):
        """
        Return an iterable of the times at which the given item is scheduled to
        run.
        """
        events = self.store.query(
            TimedEvent, TimedEvent.runnable == runnable)
        return (event.time for event in events if not event.running)

_EPSILON = 1e-20      # A very small amount of time.



class _SiteScheduler(SchedulerMixin, Service, object):
    """
    Adapter from a site store to L{IScheduler}.
    """
    implements(IScheduler)

    timer = None
    callLater = reactor.callLater
    now = Time

    def __init__(self, store):
        self.store = store
        self.setName(SITE_SCHEDULER)


    def startService(self):
        """
        Start calling persistent timed events whose time has come.
        """
        super(_SiteScheduler, self).startService()
        self._transientSchedule(self.now(), self.now())


    def stopService(self):
        """
        Stop calling persistent timed events.
        """
        super(_SiteScheduler, self).stopService()
        if self.timer is not None:
            self.timer.cancel()
            self.timer = None


    def tick(self):
        self.timer = None
        return super(_SiteScheduler, self).tick()


    def _transientSchedule(self, when, now):
        """
        If the service is currently running, schedule a tick to happen no
        later than C{when}.

        @param when: The time at which to tick.
        @type when: L{epsilon.extime.Time}

        @param now: The current time.
        @type now: L{epsilon.extime.Time}
        """
        if not self.running:
            return
        if self.timer is not None:
            if self.timer.getTime() < when.asPOSIXTimestamp():
                return
            self.timer.cancel()
        delay = when.asPOSIXTimestamp() - now.asPOSIXTimestamp()

        # reactor.callLater allows only positive delay values.  The scheduler
        # may want to have scheduled things in the past and that's OK, since we
        # are dealing with Time() instances it's impossible to predict what
        # they are relative to the current time from user code anyway.
        delay = max(_EPSILON, delay)
        self.timer = self.callLater(delay, self.tick)
        self.nextEventAt = when



class _UserScheduler(SchedulerMixin, Service, object):
    """
    Adapter from a non-site store to L{IScheduler}.
    """
    implements(IScheduler)

    def __init__(self, store):
        self.store = store


    def now(self):
        """
        Report the current time, as reported by the parent's scheduler.
        """
        return IScheduler(self.store.parent).now()


    def _transientSchedule(self, when, now):
        """
        If this service's store is attached to its parent, ask the parent to
        schedule this substore to tick at the given time.

        @param when: The time at which to tick.
        @type when: L{epsilon.extime.Time}

        @param now: Present for signature compatibility with
            L{_SiteScheduler._transientSchedule}, but ignored otherwise.
        """
        if self.store.parent is not None:
            subStore = self.store.parent.getItemByID(self.store.idInParent)
            hook = self.store.parent.findOrCreate(
                _SubSchedulerParentHook,
                subStore=subStore)
            hook._schedule(when)


    def migrateDown(self):
        """
        Remove the components in the site store for this SubScheduler.
        """
        subStore = self.store.parent.getItemByID(self.store.idInParent)
        ssph = self.store.parent.findUnique(
            _SubSchedulerParentHook,
            _SubSchedulerParentHook.subStore == subStore,
            default=None)
        if ssph is not None:
            te = self.store.parent.findUnique(TimedEvent,
                                              TimedEvent.runnable == ssph,
                                              default=None)
            if te is not None:
                te.deleteFromStore()
            ssph.deleteFromStore()


    def migrateUp(self):
        """
        Recreate the hooks in the site store to trigger this SubScheduler.
        """
        te = self.store.findFirst(TimedEvent, sort=TimedEvent.time.descending)
        if te is not None:
            self._transientSchedule(te.time, None)



class _SchedulerCompatMixin(object):
    """
    Backwards compatibility helper for L{Scheduler} and L{SubScheduler}.

    This mixin provides all the attributes from L{IScheduler}, but provides
    them by adapting the L{Store} the item is in to L{IScheduler} and
    getting them from the resulting object.  Primarily in support of test
    code, it also supports rebinding those attributes by rebinding them on
    the L{IScheduler} powerup.

    @see: L{IScheduler}
    """
    implements(IScheduler)

    def forwardToReal(name):
        def get(self):
            return getattr(IScheduler(self.store), name)
        def set(self, value):
            setattr(IScheduler(self.store), name, value)
        return property(get, set)

    now = forwardToReal("now")
    tick = forwardToReal("tick")
    schedule = forwardToReal("schedule")
    reschedule = forwardToReal("reschedule")
    unschedule = forwardToReal("unschedule")
    unscheduleAll = forwardToReal("unscheduleAll")
    scheduledTimes = forwardToReal("scheduledTimes")


    def activate(self):
        """
        Whenever L{Scheduler} or L{SubScheduler} is created, either newly or
        when loaded from a database, emit a deprecation warning referring
        people to L{IScheduler}.
        """
        # This is unfortunate.  Perhaps it is the best thing which works (it is
        # the first I found). -exarkun
        if '_axiom_memory_dummy' in vars(self):
            stacklevel = 7
        else:
            stacklevel = 5
        warnings.warn(
            self.__class__.__name__ + " is deprecated since Axiom 0.5.32.  "
            "Just adapt stores to IScheduler.",
            category=PendingDeprecationWarning,
            stacklevel=stacklevel)



class Scheduler(Item, _SchedulerCompatMixin):
    """
    Track and execute persistent timed events for a I{site} store.

    This is deprecated and present only for backwards compatibility.  Adapt
    the store to L{IScheduler} instead.
    """
    implements(IService)

    typeName = 'axiom_scheduler'
    schemaVersion = 2

    dummy = integer()

    def activate(self):
        _SchedulerCompatMixin.activate(self)


    def setServiceParent(self, parent):
        """
        L{Scheduler} is no longer an L{IService}, but still provides this
        method as a no-op in case an instance which was still an L{IService}
        powerup is loaded (in which case it will be used like a service
        once).
        """



declareLegacyItem(
    Scheduler.typeName, 1,
    dict(eventsRun=integer(default=0),
         lastEventAt=timestamp(),
         nextEventAt=timestamp()))


def scheduler1to2(old):
    new = old.upgradeVersion(Scheduler.typeName, 1, 2)
    new.store.powerDown(new, IService)
    new.store.powerDown(new, IScheduler)
    return new

registerUpgrader(scheduler1to2, Scheduler.typeName, 1, 2)


class _SubSchedulerParentHook(Item):
    schemaVersion = 4
    typeName = 'axiom_subscheduler_parent_hook'

    subStore = reference(
        doc="""
        The L{SubStore} for which this scheduling hook exists.
        """, reftype=SubStore)

    def run(self):
        """
        Tick our C{subStore}'s L{SubScheduler}.
        """
        IScheduler(self.subStore).tick()


    def _schedule(self, when):
        """
        Ensure that this hook is scheduled to run at or before C{when}.
        """
        sched = IScheduler(self.store)
        for scheduledAt in sched.scheduledTimes(self):
            if when < scheduledAt:
                sched.reschedule(self, scheduledAt, when)
            break
        else:
            sched.schedule(self, when)


def upgradeParentHook1to2(oldHook):
    """
    Add the scheduler attribute to the given L{_SubSchedulerParentHook}.
    """
    newHook = oldHook.upgradeVersion(
        oldHook.typeName, 1, 2,
        loginAccount=oldHook.loginAccount,
        scheduledAt=oldHook.scheduledAt,
        scheduler=oldHook.store.findFirst(Scheduler))
    return newHook

registerUpgrader(upgradeParentHook1to2, _SubSchedulerParentHook.typeName, 1, 2)

declareLegacyItem(
    _SubSchedulerParentHook.typeName, 2,
    dict(loginAccount=reference(),
         scheduledAt=timestamp(default=None),
         scheduler=reference()))

def upgradeParentHook2to3(old):
    """
    Copy the C{loginAccount} attribute, but drop the others.
    """
    return old.upgradeVersion(
        old.typeName, 2, 3,
        loginAccount=old.loginAccount)

registerUpgrader(upgradeParentHook2to3, _SubSchedulerParentHook.typeName, 2, 3)

declareLegacyItem(
    _SubSchedulerParentHook.typeName, 3,
    dict(loginAccount=reference(),
         scheduler=reference()))

def upgradeParentHook3to4(old):
    """
    Copy C{loginAccount} to C{subStore} and remove the installation marker.
    """
    new = old.upgradeVersion(
        old.typeName, 3, 4, subStore=old.loginAccount)
    uninstallFrom(new, new.store)
    return new


registerUpgrader(upgradeParentHook3to4, _SubSchedulerParentHook.typeName, 3, 4)


class SubScheduler(Item, _SchedulerCompatMixin):
    """
    Track and execute persistent timed events for a substore.

    This is deprecated and present only for backwards compatibility.  Adapt
    the store to L{IScheduler} instead.
    """
    schemaVersion = 2
    typeName = 'axiom_subscheduler'

    dummy = integer()

    def activate(self):
        _SchedulerCompatMixin.activate(self)


def subscheduler1to2(old):
    new = old.upgradeVersion(SubScheduler.typeName, 1, 2)
    try:
        new.store.powerDown(new, IScheduler)
    except ValueError:
        # Someone might have created a SubScheduler but failed to power it
        # up.  Fine.
        pass
    return new

registerUpgrader(subscheduler1to2, SubScheduler.typeName, 1, 2)
